from collections import defaultdict
from typing import Any, NamedTuple


# ----------------------------------------------------------------------------
# KEY: Set dictionary keys on a per-os_type basis
# ----------------------------------------------------------------------------
class OsKeys(NamedTuple):
    nve_vni_bd_vrf: str


def _set_keys(os_type: str) -> OsKeys:
    """Based on the OS type set the set the dictionary keys when gleaning data from NTC data structures.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
    Returns:
        OsKeys: Dictionary Keys for the specific OS type to retrieve the output data
    """
    if "ios" in os_type:
        return OsKeys("vrf")
    elif "nxos" in os_type:
        return OsKeys("bd_vrf")
    # Fallback if nothing matched
    msg = f"Error, '_set_keys' has no match for OS type: '{os_type}'"
    raise NotImplementedError(msg)


# ----------------------------------------------------------------------------
# OUTPUT: Creates str or ntc output dictionaries based on device command output
# ----------------------------------------------------------------------------
def _format_output(
    os_type: str, sub_feature: str, output: list[str | dict[str, str]]
) -> tuple[list[str], list[dict[str, str]]]:
    """Screen scraping return different data structures, they need defining to make function typing easier.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str from raw) command output from the device
    Raises:
        ValueError: Errors if the new output variable doesn't match the input as not meant to be changing it, just defining it for MYPY
    Returns:
        tuple[list[str], list[dict[str, str]]]: Returns either RAW list[str] or NTC list[dict], the other non-matched one will be empty
    """
    raw_output = [o for o in output if isinstance(o, str)]
    ntc_output = [o for o in output if isinstance(o, dict)]
    if (
        len(raw_output) != 0
        and output != raw_output
        or len(ntc_output) != 0
        and output != ntc_output
    ):
        msg = f"{os_type} {sub_feature} _format_output is malformed"
        raise ValueError(msg)
    return raw_output, ntc_output


# ----------------------------------------------------------------------------
# DEF: Mini-functions used by the main function
# ----------------------------------------------------------------------------
def _make_int(input_data: str) -> int | str:
    """Takes a string and returns an integer if it can, otherwise it returns the original string.

    Args:
        input_data (str): The data to be converted to an integer
    Returns:
        int | str: The input_data as a integer if possible, if not as the original string
    """
    try:
        return int(input_data)
    except ValueError:
        return input_data


def format_nve_vni(
    val_file: bool, key: OsKeys, output: list[dict[str, str]]
) -> dict[str | int, Any]:
    """Format NVE VNI into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        output (list[dict[str, str]]): The command output from the device in ntc data structure
    Returns:
        dict[str | int, Any]: {l3vni: {bdi_vrf: z, state: Up}}, val file doesn't include state
    """
    result: dict[str | int, dict[str, str | int]] = defaultdict(dict)
    for each_vni in output:
        vni = _make_int(each_vni["vni"])
        if "L2" in each_vni["mode"]:
            result[vni]["bd_vrf"] = _make_int(each_vni["bd"])
        else:
            result[vni]["bd_vrf"] = _make_int(each_vni[key.nve_vni_bd_vrf])
        # If creating actual state
        if not val_file:
            result[vni]["state"] = each_vni["state"]
    return dict(result)


def format_nve_peer(
    val_file: bool, output: list[dict[str, str]]
) -> dict[str, str] | list[str]:
    """Format NVE VNI into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        output (list[dict[str, str]]): The command output from the device in ntc data structure
    Returns:
        dict[str, str] | list[str]: {peer1_ip: state: Up}, val file is just list a [peer_ip, peer_ip]
    """
    result = {}
    val_file_list: list[str] = []
    for each_peer in output:
        result[each_peer["peer"]] = each_peer["state"].capitalize()
        if each_peer["peer"] not in str(val_file_list):
            val_file_list.append(each_peer["peer"])
    # List of peers if is a validation file
    if val_file:
        return val_file_list
    else:
        return result


# ----------------------------------------------------------------------------
# ACTUAL_STATE: Engine use to create sub-feature actual state or validation file
# ----------------------------------------------------------------------------
def format_actual_state(
    val_file: bool,  # noqa: ARG001
    os_type: str,
    sub_feature: str,
    output: list[str | dict[str, str]],
) -> dict[Any, Any] | list[str]:
    """Engine to run all the actual state and validation file sub-feature formatting.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str/int from raw) command output from the device
    Returns:
        dict[Any, Any] | list[str]: Returns cmd output formatted into the data structure of actual state or validation file
    """
    key = _set_keys(os_type)
    raw_output, ntc_output = _format_output(os_type, sub_feature, output)

    ### NVE_VNI: {l3vni: {bdi_vrf: z, state: Up}}, val file doesn't include state
    if sub_feature == "nve_vni":
        return format_nve_vni(val_file, key, ntc_output)
    ### NVE_PEER: {peer1_ip: state: Up}, val file is just list a [peer_ip, peer_ip]
    elif sub_feature == "nve_peer":
        return format_nve_peer(val_file, ntc_output)
    ### CatchAll
    else:
        msg = f"Unsupported sub_feature: {sub_feature}"
        raise ValueError(msg)



from collections import defaultdict
from typing import Any, NamedTuple


# ----------------------------------------------------------------------------
# KEY: Set dictionary keys on a per-os_type basis
# ----------------------------------------------------------------------------
class OsKeys(NamedTuple):
    mac_table_match: str
    mac_table_element: int
    mac_table_idx: int


def _set_keys(os_type: str) -> OsKeys:
    """Based on the OS type set the set the dictionary keys when gleaning data from NTC data structures.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
    Returns:
        OsKeys: Dictionary Keys for the specific OS type to retrieve the output data
    """
    if "ios" in os_type:
        return OsKeys("Vlan", -1, 1)
    elif "nxos" in os_type:
        return OsKeys("enet", 0, 1)
    # Fallback if nothing matched
    msg = f"Error, '_set_keys' has no match for OS type: '{os_type}'"
    raise NotImplementedError(msg)


# ----------------------------------------------------------------------------
# OUTPUT: Creates str or ntc output dictionaries based on device command output
# ----------------------------------------------------------------------------
def _format_output(
    os_type: str, sub_feature: str, output: list[str | dict[str, str]]
) -> tuple[list[str], list[dict[str, str]]]:
    """Screen scraping return different data structures, they need defining to make function typing easier.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str from raw) command output from the device
    Raises:
        ValueError: Errors if the new output variable doesn't match the input as not meant to be changing it, just defining it for MYPY
    Returns:
        tuple[list[str], list[dict[str, str]]]: Returns either RAW list[str] or NTC list[dict], the other non-matched one will be empty
    """
    raw_output = [str(o) for o in output if isinstance(o, (str, int))]
    ntc_output = [o for o in output if isinstance(o, dict)]
    if len(raw_output) != 0 and output != raw_output:
        pass  # Pass instead of RAISE as it changes all int into str (makes MYPY checks easier)
    elif len(ntc_output) != 0 and output != ntc_output:
        msg = f"{os_type} {sub_feature} _format_output is malformed"
        raise ValueError(msg)
    return raw_output, ntc_output


# ----------------------------------------------------------------------------
# DEF: Mini-functions used by the main function
# ----------------------------------------------------------------------------
def _make_int(input_data: str) -> int | str:
    """Takes a string and returns an integer if it can, otherwise it returns the original string.

    Args:
        input_data (str): The data to be converted to an integer
    Returns:
        int | str: The input_data as a integer if possible, if not as the original string
    """
    try:
        return int(input_data)
    except ValueError:
        return input_data


def format_vlan(output: list[dict[str, Any]]) -> dict[str | int, Any]:
    """Format vlan output into the data structure.

    Args:
        output (list[dict[str, Any]]): The command output from the device in ntc data structure
    Returns:
        dict[str | int, Any]: {vlan: {name: x, intf:[x,y]}}
    """
    result: dict[str | int, dict[str, dict[str, list[str]]]] = defaultdict(dict)
    for each_vl in output:
        vl_id = _make_int(each_vl["vlan_id"])
        result[vl_id]["name"] = each_vl["vlan_name"]
        result[vl_id]["intf"] = each_vl["interfaces"]
    for each_item in [1002, 1003, 1004, 1005]:
        if result.get(each_item) is not None:
            del result[each_item]
    return dict(result)


def format_stp(val_file: bool, output: list[dict[str, Any]]) -> dict[str | int, Any]:
    """Format STP into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        output (list[dict[str, Any]]): The command output from the device in ntc data structure
    Returns:
        dict[str | int, Any]: {vlan: {intfx: FWD, intfy: FWD}}, if val_file is {vlan: {'intf': [intfx, intfy]}}
    """
    result: dict[str | int, dict[str, Any]] = defaultdict(dict)
    for each_vl in output:
        # If creating actual state
        if not val_file:
            result[_make_int(each_vl["vlan_id"])][each_vl["interface"]] = each_vl[
                "status"
            ]
        # If creating validation file
        elif val_file:
            if result.get(_make_int(each_vl["vlan_id"])) is None:
                result[_make_int(each_vl["vlan_id"])]["intf"] = [each_vl["interface"]]
            elif isinstance(result[_make_int(each_vl["vlan_id"])]["intf"], list):
                result[_make_int(each_vl["vlan_id"])]["intf"].append(
                    each_vl["interface"]
                )
    return dict(result)


def format_mac_table(key: OsKeys, output: list[str]) -> dict[str, str | int]:
    """Format MAC table count into the data structure.

    Args:
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        output (list[str]): The command output from the device raw data structure
    Results:
        dict[str, str | int]: {total_mac_count: x, vlx_mac_count: x}
    """
    result = {}
    # Total MAC count is always first element in list
    try:
        result["total_mac_count"] = _make_int(output[0].split()[-1])
    except Exception as e:
        result["total_mac_count"] = 0
    # Per-VLAN MAC count are in 2 consecutive cmds, first VL to match on and second the count
    for idx, each_item in enumerate(output):
        each_item = str(each_item)
        if key.mac_table_match in each_item:
            name = f"vl{each_item.split()[key.mac_table_element].replace(':', '')}_mac_count"
            result[name] = _make_int(output[idx + key.mac_table_idx].split()[-1])
            output[idx] = ""
            output[idx + 1] = ""
    return dict(result)


# ----------------------------------------------------------------------------
# ACTUAL_STATE: Engine use to create sub-feature actual state or validation file
# ----------------------------------------------------------------------------
def format_actual_state(
    val_file: bool,
    os_type: str,
    sub_feature: str,
    output: list[str | dict[str, str]],
) -> dict[Any, Any]:
    """Engine to run all the actual state and validation file sub-feature formatting.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str/int from raw) command output from the device
    Returns:
        dict[Any, Any]: Returns cmd output formatted into the data structure of actual state or validation file
    """
    key = _set_keys(os_type)
    raw_output, ntc_output = _format_output(os_type, sub_feature, output)

    ### VLAN: {vlan: {name: x, intf:[x,y]}}
    if sub_feature == "vlan":
        return format_vlan(ntc_output)

    ### STP: {vlan: {intfx: FWD, intfy: FWD}}, if val_file is {vlan: {'intf': [intfx, intfy]}}
    elif sub_feature == "stp_vlan":
        return format_stp(val_file, ntc_output)

    ### MAC TABLE COUNT: {total_mac_count: x, vlx_mac_count: x}
    elif sub_feature == "mac_table":
        return format_mac_table(key, raw_output)

    ### CatchAll
    else:
        msg = f"Unsupported sub_feature: {sub_feature}"
        raise ValueError(msg)



import re
from collections import defaultdict
from typing import Any, NamedTuple


# ----------------------------------------------------------------------------
# KEY: Set dictionary keys on a per-os_type basis
# ----------------------------------------------------------------------------
class OsKeys(NamedTuple):
    hsrp_intf: str
    hsrp_prio: str
    hsrp_state: str


def _set_keys(os_type: str) -> OsKeys:
    """Based on the OS type set the set the dictionary keys when gleaning data from NTC data structures.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
    Returns:
        OsKeys: A NamedTuple with the keys for interface, priority, and state
    """
    if "ios" in os_type:
        return OsKeys("interface", "priority", "state")
    elif "nxos" in os_type:
        return OsKeys("sh_if_index", "sh_prio", "sh_group_state")
    elif bool(re.search("asa", os_type)):  # noqa: SIM114
        pass
    elif bool(re.search("wlc", os_type)):
        pass
    # Fallback if nothing matched
    msg = f"Error, '_set_keys' has no match for OS type: '{os_type}'"
    raise NotImplementedError(msg)


# ----------------------------------------------------------------------------
# OUTPUT: Creates str or ntc output dictionaries based on device command output
# ----------------------------------------------------------------------------
def _format_output(
    os_type: str, sub_feature: str, output: list[str | dict[str, str]]
) -> tuple[list[str], list[dict[str, str]]]:
    """Screen scraping return different data structures, they need defining to make function typing easier.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str from raw) command output from the device
    Raises:
        ValueError: Errors if the new output variable doesn't match the input as not meant to be changing it, just defining it for MYPY
    Returns:
        tuple[list[str], list[dict[str, str]]]: Returns either RAW list[str] or NTC list[dict], the other non-matched one will be empty
    """
    raw_output = [o for o in output if isinstance(o, str)]
    ntc_output = [o for o in output if isinstance(o, dict)]
    if (
        len(raw_output) != 0
        and output != raw_output
        or len(ntc_output) != 0
        and output != ntc_output
    ):
        msg = f"{os_type} {sub_feature} _format_output is malformed"
        raise ValueError(msg)
    return raw_output, ntc_output


# ----------------------------------------------------------------------------
# DEF: Mini-functions used by the main function
# ----------------------------------------------------------------------------
def _make_int(input_data: str) -> int | str:
    """Takes a string and returns an integer if it can, otherwise it returns the original string.

    Args:
        input_data (str): The data to be converted to an integer
    Returns:
        int | str: The input_data as a integer if possible, if not as the original string
    """
    try:
        return int(input_data)
    except ValueError:
        return input_data


def _fix_nxos(
    main_dict: dict[str, Any], parent_name: str, child_name: str
) -> list[dict[str, Any]]:
    """NXOS JSON makes dict rather than list if only 1 item in output. If the child_dict is a dict this methods convert it to a list.

    Args:
        main_dict (dict[str, Any]): The dictionary of the command output
        parent_name (str): The parent dictionary key name (e.g. TABLE_xx)
        child_name (str): The key of the dictionary that you want to fix, for example ROW_xx
    Returns:
        A list of dictionaries
    """
    child_dict_value: list[dict[str, Any]] | dict[str, Any] = main_dict[parent_name][
        child_name
    ]
    if isinstance(child_dict_value, dict):
        child_dict_value = [child_dict_value]
    return child_dict_value


def format_hsrp(
    key: OsKeys, output: list[dict[str, Any]]
) -> dict[str, dict[str, str | int]]:
    """Format HSRP output into the data structure.

    Args:
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        output (list[dict[str, Any]]): The command output from the device in ntc data structure
    Returns:
        dict[str, dict[str, str | int]]: {intf: {priority: x, state: y}
    """
    result: dict[str, dict[str, str | int]] = defaultdict(dict)
    for each_nbr in output:
        result[each_nbr[key.hsrp_intf]]["priority"] = _make_int(each_nbr[key.hsrp_prio])
        result[each_nbr[key.hsrp_intf]]["state"] = each_nbr[key.hsrp_state]
    return dict(result)


# ----------------------------------------------------------------------------
# ACTUAL_STATE: Engine use to create sub-feature actual state or validation file
# ----------------------------------------------------------------------------
def format_actual_state(
    val_file: bool,  # noqa: ARG001
    os_type: str,
    sub_feature: str,
    output: list[str | dict[str, str]],
) -> dict[str, dict[str, str | int]]:
    """Engine to run all the actual state and validation file sub-feature formatting.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str/int from raw) command output from the device
    Returns:
        dict[str, dict[str, str | int]]: Returns cmd output formatted into the data structure of actual state or validation file
    """
    key = _set_keys(os_type)
    raw_output, ntc_output = _format_output(os_type, sub_feature, output)

    ### HSRP: {intf: {priority: x, state: y})
    if sub_feature == "hsrp":  # noqa: SIM102
        if bool(re.search("nxos", os_type)):
            ntc_output = _fix_nxos(ntc_output[0], "TABLE_grp_detail", "ROW_grp_detail")
        return format_hsrp(key, ntc_output)

    ### CatchAll
    else:
        msg = f"Unsupported sub_feature: {sub_feature}"
        raise ValueError(msg)



import re
from collections import Counter, defaultdict
from typing import Any, NamedTuple


# ----------------------------------------------------------------------------
# KEY: Set dictionary keys on a per-os_type basis
# ----------------------------------------------------------------------------
class OsKeys(NamedTuple):
    neighbor_id: str
    bgp_nhbr: str
    bgp_nhbr_as: str
    bgp_pfxrcd: str


def _set_keys(os_type: str) -> OsKeys:
    """Based on the OS type set the set the dictionary keys when gleaning data from NTC data structures.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
    Returns:
        OsKeys: Dictionary Keys for the specific OS type to retrieve the output data
    """
    if "ios" in os_type:
        return OsKeys(
            "neighbor_id", "bgp_neighbor", "neighbor_as", "state_or_prefixes_received"
        )
    elif "nxos" in os_type:  # noqa: SIM114
        return OsKeys("neighbor_id", "bgp_neigh", "neigh_as", "state_pfxrcd")
    elif "asa" in os_type:
        return OsKeys("neighbor_id", "bgp_neigh", "neigh_as", "state_pfxrcd")
    elif "panos" in os_type:
        return OsKeys("", "bgp_neighbor", "neighbor_as", "accepted_pfx")
    # Fallback if nothing matched
    msg = f"Error, '_set_keys' has no match for OS type: '{os_type}'"
    raise NotImplementedError(msg)


# ----------------------------------------------------------------------------
# OUTPUT: Creates str or ntc output dictionaries based on device command output
# ----------------------------------------------------------------------------
def _format_output(
    os_type: str, sub_feature: str, output: list[str | dict[str, str]]
) -> tuple[list[str], list[dict[str, str]]]:
    """Screen scraping return different data structures, they need defining to make function typing easier.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str from raw) command output from the device
    Raises:
        ValueError: Errors if the new output variable doesn't match the input as not meant to be changing it, just defining it for MYPY
    Returns:
        tuple[list[str], list[dict[str, str]]]: Returns either RAW list[str] or NTC list[dict], the other non-matched one will be empty
    """
    raw_output = [o for o in output if isinstance(o, str)]
    ntc_output = [o for o in output if isinstance(o, dict)]
    if (
        len(raw_output) != 0
        and output != raw_output
        or len(ntc_output) != 0
        and output != ntc_output
    ):
        msg = f"{os_type} {sub_feature} _format_output is malformed"
        raise ValueError(msg)
    return raw_output, ntc_output


# ----------------------------------------------------------------------------
# DEF: Mini-functions used by the main function
# ----------------------------------------------------------------------------
def _make_int(input_data: str) -> int | str:
    """Takes a string and returns an integer if it can, otherwise it returns the original string.

    Args:
        input_data (str): The data to be converted to an integer
    Returns:
        int | str: The input_data as a integer if possible, if not as the original string
    """
    try:
        return int(input_data)
    except ValueError:
        return input_data


def format_eigrp_nhbr(val_file: bool, output: list[dict[str, Any]]) -> dict[str, Any]:
    """Format EIGRP neighbor output into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        output (list[dict[str, str]]): The command output from the device in ntc data structure
    Returns:
        dict[str, Any]: {intf: {asn: x, nbr: {nbr_ip: up}}, val_file is {intf: {asn: x, nbr: nbr_ip}}
    """
    result: dict[str, dict[str, str | int | dict[str, str]]] = defaultdict(dict)
    for each_nbr in output:
        intf = each_nbr["interface"]
        state = "down" if len(each_nbr["uptime"]) == 0 else "up"
        if result.get(intf) is None:
            result[intf]["asn"] = _make_int(each_nbr["as"])
        # val_file
        if val_file:
            result[intf]["nbr"] = each_nbr["ip_address"]
        # actual state
        elif not val_file:
            result[intf]["nbr"] = {each_nbr["ip_address"]: state}
    return dict(result)


def format_ospf_nhbr(
    val_file: bool, os_type: str, key: OsKeys, output: list[dict[str, Any]]
) -> dict[str, Any]:
    """Format OSPF interface and neighbor outputs into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        output (list[dict[str, str]]): The command output from the device in ntc data structure
    Returns:
        dict[str, Any]: {intf: {pid, x, area: x, nbr: {nbr_ip: FULL}}, val file is {intf: {pid, x, area: x, nbr: [nbr_ip, nbr_ip]}}
    """
    result: dict[str, Any] = defaultdict(dict)

    # Helper functions used only by this function
    def _remove_char(input_data: str, char: str) -> str:
        """If specified character the input data, return the input data split on the specified character, otherwise return the input data."""
        if char in input_data:
            return input_data.split(char)[0]
        else:
            return input_data

    def _ospf_nhbr(
        neighbor_id: str,
        output: list[dict[str, Any]],
        tmp_dict_nbr: defaultdict[str, Any],
    ) -> dict[str, list[str] | dict[str, str | int | dict[str, str]]]:
        """Returns formatted OSPF neighbors combined and grouped by interface."""
        for each_intf in output:
            # Data from "show ip int br" is saved in result
            if each_intf.get("process") is not None:
                pid = _make_int(each_intf["process"])
                result[each_intf["interface"]]["pid"] = pid
                result[each_intf["interface"]]["area"] = _make_int(each_intf["area"])
                # If is actual state adds empty dict
                if tmp_dict_nbr.default_factory is dict:
                    result[each_intf["interface"]]["nbr"] = {}
            # Data from "show ip ospf neigh" is saved in tmp_dict_nbr (list is val_file, dict actual_state)
            elif tmp_dict_nbr.default_factory is list:
                tmp_dict_nbr[each_intf["interface"]].extend([each_intf[neighbor_id]])
            elif tmp_dict_nbr.default_factory is dict:
                state = _remove_char(each_intf["state"], "/")
                tmp_dict_nbr[each_intf["interface"]][each_intf[neighbor_id]] = state
        return dict(tmp_dict_nbr)

    # INTF: Add OSPF intf and process/area to results and neigbors per interface to dict_nbr
    if val_file:
        dict_nbr = _ospf_nhbr(key.neighbor_id, output, defaultdict(list))
    elif not val_file:
        dict_nbr = _ospf_nhbr(key.neighbor_id, output, defaultdict(dict))
    # NBR: Add neighbors based on matching interfaces in results and dict_nbr
    for intf in result.keys():  # noqa: SIM118
        for intf1, nbr in dict_nbr.items():
            # ASA has to be treated differently as interfaces are logical names
            if bool(re.search("asa", os_type)):
                if intf == intf1:
                    result[intf]["nbr"] = nbr
            else:
                # Splits intf short and long names to then be able to interface number ([1:]) and short name is in long name ([0])
                spl_intf = re.split(r"(\D+)", intf.lower())[1:]
                spl_intf1 = re.split(r"(\D+)", intf1.lower())[1:]
                if spl_intf[1:] == spl_intf1[1:] and spl_intf[0] in spl_intf1[0]:
                    result[intf]["nbr"] = nbr
    return dict(result)


def format_ospf_lsdb(output: list[str]) -> dict[str | int, Any]:
    """Format OSPF LSDB count output into the data structure.

    Args:
        output (list[dict[str, str]]): The command output from the device in raw data structure
    Returns:
        ict[str | int, Any]: {process: total_lsa: x}
    """
    result: dict[str | int, dict[str, str | int]] = {}
    for idx, each_item in enumerate(output):
        if "Process ID" in each_item:
            proc = _make_int(each_item.split("s ID")[1].replace(")", "").split()[0])
            result[proc] = {"total_lsa": _make_int(output[idx + 1].split()[1])}
    return result


def format_bgp(
    os_type: str, key: OsKeys, output: list[dict[str, Any]]
) -> dict[str, Any]:
    """Format BGP peers into the data structure.

    Args:
        os_type (str): The different Nornir platforms which are OS type of the device
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        output (list[dict[str, str]]): The command output from the device in ntc data structure
    Returns:
        dict[str, Any]: {peer: {asn:x, rcv_pfx:x}}
    """
    result: dict[str, Any] = defaultdict(dict)

    # Helper functions used only by this function
    def _remove_dup_peers(
        output: list[dict[str, Any]], bgp_nhbr: str
    ) -> list[dict[str, str]]:
        """Removes duplicate peers from the IPv4 address family as likely to be the underlay peers."""
        neigh_counts = Counter(item[bgp_nhbr] for item in output)
        cln_output = [
            item
            for item in output
            if not (
                neigh_counts[item[bgp_nhbr]] > 1
                and item["address_family"] == "IPv4 Unicast"
            )
        ]
        return cln_output

    if not bool(re.search("asa", os_type)):
        cln_output = _remove_dup_peers(output, key.bgp_nhbr)
    else:
        cln_output = output
    for each_peer in cln_output:
        result[each_peer[key.bgp_nhbr]]["asn"] = _make_int(each_peer[key.bgp_nhbr_as])
        result[each_peer[key.bgp_nhbr]]["rcv_pfx"] = _make_int(
            each_peer[key.bgp_pfxrcd]
        )
    return dict(result)


# ----------------------------------------------------------------------------
# ACTUAL_STATE: Engine use to create sub-feature actual state or validation file
# ----------------------------------------------------------------------------
def format_actual_state(
    val_file: bool,
    os_type: str,
    sub_feature: str,
    output: list[str | dict[str, str]],
) -> dict[Any, Any]:
    """Engine to run all the actual state and validation file sub-feature formatting.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str/int from raw) command output from the device
    Returns:
        dict[str | int, Any]: Returns cmd output formatted into the data structure of actual state or validation file
    """
    key = _set_keys(os_type)
    raw_output, ntc_output = _format_output(os_type, sub_feature, output)

    ### EIGRP_INTF_NBR: {intf: {asn: x, nbr: {nbr_ip: up}}
    if sub_feature == "eigrp_intf_nbr":
        return format_eigrp_nhbr(val_file, ntc_output)

    ### OSPF_INTF_NBR: {intf: {pid, x, area: x, nbr: {nbr_ip: FULL}}
    elif sub_feature == "ospf_intf_nbr":
        return format_ospf_nhbr(val_file, os_type, key, ntc_output)

    ### OSPF_LSDB_COUNT: {process: total_lsa: x}
    elif sub_feature == "ospf_lsdb_count":
        return format_ospf_lsdb(raw_output)

    ### BGP_PEER: {peer: {asn:x, rcv_pfx:x}}
    elif sub_feature == "bgp_peer":
        return format_bgp(os_type, key, ntc_output)

    ### CatchAll
    else:
        msg = f"Unsupported sub_feature: {sub_feature}"
        raise ValueError(msg)



import re
from collections import defaultdict
from typing import Any, NamedTuple


# ----------------------------------------------------------------------------
# KEY: Set dictionary keys on a per-os_type basis
# ----------------------------------------------------------------------------
class OsKeys(NamedTuple):
    po_name: str
    po_status: str
    po_protocol: str
    po_intf: str


def _set_keys(os_type: str) -> OsKeys:
    """Based on the OS type set the set the dictionary keys when gleaning data from NTC data structures.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
    Returns:
        OsKeys: Dictionary Keys for the specific OS type to retrieve the output data
    """
    if "ios" in os_type or "nxos" in os_type or "asa" in os_type:
        return OsKeys(
            "bundle_name", "bundle_status", "bundle_protocol", "member_interface"
        )
    # Fallback if nothing matched
    msg = f"Error, '_set_keys' has no match for OS type: '{os_type}'"
    raise NotImplementedError(msg)


# ----------------------------------------------------------------------------
# OUTPUT: Creates str or ntc output dictionaries based on device command output
# ----------------------------------------------------------------------------
def _format_output(
    os_type: str, sub_feature: str, output: list[str | dict[str, str]]
) -> tuple[list[str], list[dict[str, str]]]:
    """Screen scraping return different data structures, they need defining to make function typing easier.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str from raw) command output from the device
    Raises:
        ValueError: Errors if the new output variable doesn't match the input as not meant to be changing it, just defining it for MYPY
    Returns:
        tuple[list[str], list[dict[str, str]]]: Returns either RAW list[str] or NTC list[dict], the other non-matched one will be empty
    """
    raw_output = [o for o in output if isinstance(o, str)]
    ntc_output = [o for o in output if isinstance(o, dict)]
    if (
        len(raw_output) != 0
        and output != raw_output
        or len(ntc_output) != 0
        and output != ntc_output
    ):
        msg = f"{os_type} {sub_feature} _format_output is malformed"
        raise ValueError(msg)
    return raw_output, ntc_output


# ----------------------------------------------------------------------------
# DEF: Mini-functions used by the main function
# ----------------------------------------------------------------------------
def _make_int(input_data: str) -> int | str:
    """Takes a string and returns an integer if it can, otherwise it returns the original string.

    Args:
        input_data (str): The data to be converted to an integer
    Returns:
        int | str: The input_data as a integer if possible, if not as the original string
    """
    try:
        return int(input_data)
    except ValueError:
        return input_data


def _fix_nxos(
    main_dict: dict[str, Any], parent_name: str, child_name: str
) -> list[dict[str, Any]]:
    """NXOS JSON makes dict rather than list if only 1 item in output. If the child_dict is a dict this methods convert it to a list.

    Args:
        main_dict (dict[str, Any]): The dictionary of the command output
        parent_name (str): The parent dictionary key name (e.g. TABLE_xx)
        child_name (str): The key of the dictionary that you want to fix, for example ROW_xx
    Returns:
        A list of dictionaries
    """
    child_dict_value: list[dict[str, Any]] | dict[str, Any] = main_dict[parent_name][
        child_name
    ]
    if isinstance(child_dict_value, dict):
        child_dict_value = [child_dict_value]
    return child_dict_value


def _format_status(status: str) -> str:
    """Takes the status and for NXOS removes additional letters and brackets and returns a string.

    Args:
        status (str): The status of the port-channel or PO member interface
    Returns:
        str: Returns cleaned up status without any additional letters and brackets
    """
    if re.match(f".*U", status):
        status = "U"
    return status.replace("(", "").replace(")", "")


def format_po(
    val_file: bool, key: OsKeys, output: list[dict[str, Any]]
) -> dict[str, Any]:
    """Format port-channel into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        output (list[dict[str, Any]]): The command output from the device in ntc data structure
    Returns:
        dict[str, Any]: {po_name: {protocol: type, status: code, members: {intf_name: {mbr_status: code}}}}, val_file is {po_name: {protocol: type, members: [intf_x, intf_y]}}
    """
    result: dict[str, dict[str, str | dict[str, dict[str, str]]]] = defaultdict(dict)
    for each_po in output:
        if each_po[key.po_protocol] == "-":
            each_po[key.po_protocol] = "NONE"
        result[each_po[key.po_name]]["protocol"] = each_po[key.po_protocol].upper()
        # Val_file doesn't get status of PO or its member ports (is just a list of them)
        if val_file:
            result[each_po[key.po_name]]["members"] = each_po[key.po_intf]
        # The actual_state gets the PO status as well as member port status
        if not val_file:
            result[each_po[key.po_name]]["status"] = _format_status(
                each_po[key.po_status]
            )
            po_mbrs = {}
            for mbr_intf, mbr_status in zip(
                each_po[key.po_intf], each_po[f"{key.po_intf}_status"], strict=False
            ):
                # For routers as use bndl rather than P
                if mbr_status == "bndl":
                    mbr_status = "P"
                # Creates dict of members to add to as value in the PO dictionary
                po_mbrs[mbr_intf] = {"mbr_status": _format_status(mbr_status)}
            result[each_po[key.po_name]]["members"] = po_mbrs
    return dict(result)


def format_vpc_peer(val_file: bool, output: list[dict[str, Any]]) -> dict[str, Any]:
    """Format vpc_peer link info into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        output (list[dict[str, Any]]): The command output from the device in ntc data structure
    Returns:
        dict[str, Any]: {role: x, 'peerlink_vlans': [x,y], 'peerlink_port_state': x, 'peer_status': x, 'keepalive_status': x, 'vlan_consistency': x, 'peer_consistency': x, 'type2_consistency': x},
        val_file is just {role: x, 'peerlink_vlans': [x,y]}
    """
    result: dict[str, str | int | list[str | int] | dict[str, str]] = defaultdict(dict)
    # Common for both actual_state and val_file
    result["role"] = output[0]["vpc-role"]
    peer_link = output[0]["TABLE_peerlink"]["ROW_peerlink"]
    trunk_vl = peer_link["peer-up-vlan-bitset"]
    result["peerlink_vlans"] = [_make_int(vl) for vl in trunk_vl.split(",")]
    # If creating actual state also needs state (is implicit in val file)
    if not val_file:
        result["peerlink_port_state"] = _make_int(peer_link["peer-link-port-state"])
        result["peer_status"] = output[0]["vpc-peer-status"]
        result["keepalive_status"] = output[0]["vpc-peer-keepalive-status"]
        result["vlan_consistency"] = output[0]["vpc-per-vlan-peer-consistency"]
        result["peer_consistency"] = output[0]["vpc-peer-consistency-status"]
        result["type2_consistency"] = output[0]["vpc-type-2-consistency-status"]
    return dict(result)


def format_vpcs(val_file: bool, output: list[dict[str, str]]) -> dict[str | int, Any]:
    """Format vpcs into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        output (list[dict[str, str]]): The command output from the device in ntc data structure
    Returns:
        dict[str | int, Any]: {vpc_xx: {'po': x, 'vlans': [x,y], 'port_state': x, 'consistency_status': x}, val_file {vpc_xx: {'po': x, 'vlans': [x,y]}
    """
    result: dict[str | int, dict[str, str | int | list[str | int]]] = defaultdict(dict)
    all_vpcs = _fix_nxos(output[0], "TABLE_vpc", "ROW_vpc")
    for vpc in all_vpcs:
        vpc_id = _make_int(vpc["vpc-id"])
        result[vpc_id]["po"] = vpc["vpc-ifindex"]
        trunk_vl = vpc["up-vlan-bitset"]
        result[vpc_id]["vlans"] = [_make_int(vl) for vl in trunk_vl.split(",")]
        # If creating actual state also needs state (is implicit in val file)
        if not val_file:
            result[vpc_id]["port_state"] = _make_int(vpc["vpc-port-state"])
            result[vpc_id]["consistency_status"] = vpc["vpc-consistency-status"]
    return dict(result)


# ----------------------------------------------------------------------------
# ACTUAL_STATE: Engine use to create sub-feature actual state or validation file
# ----------------------------------------------------------------------------
def format_actual_state(
    val_file: bool,
    os_type: str,
    sub_feature: str,
    output: list[str | dict[str, str]],
) -> dict[str, Any]:
    """Engine to run all the actual state and validation file sub-feature formatting.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str/int from raw) command output from the device
    Returns:
        dict[str, Any]: Returns cmd output formatted into the data structure of actual state or validation file
    """
    key = _set_keys(os_type)
    aw_output, ntc_output = _format_output(os_type, sub_feature, output)

    ### PORT_CHANNEL: {po_name: {protocol: type, status: code, members: {intf_name: {mbr_status: code}}}}
    if sub_feature == "port_channel":
        return format_po(val_file, key, ntc_output)

    ### VPC: see below
    elif sub_feature == "vpc":
        # {role: x, 'peerlink_vlans': [x,y], 'peerlink_port_state': x, 'peer_status': x,'keepalive_status': x, 'vlan_consistency': x, 'peer_consistency': x, 'type2_consistency': x}, val_file is just {role: x, 'peerlink_vlans': [x,y]}
        vpc_result = format_vpc_peer(val_file, ntc_output)
        # TABLE_vpc is only present if are vPCs configured
        if ntc_output[0].get("TABLE_vpc") is not None:
            # {vpc_xx: {'po': x, 'vlans': [x,y], 'port_state': x, 'consistency_status': x}, val_file {vpc_xx: {'po': x, 'vlans': [x,y]}
            all_vpcs: dict[Any, Any] = format_vpcs(val_file, ntc_output)
            vpc_result["vpcs"] = all_vpcs
        return vpc_result

    ### CatchAll
    else:
        msg = f"Unsupported sub_feature: {sub_feature}"
        raise ValueError(msg)



import ipaddress
import re
import trace
from collections import defaultdict
from typing import Any, NamedTuple


# ----------------------------------------------------------------------------
# KEY: Set dictionary keys on a per-os_type basis
# ----------------------------------------------------------------------------
class OsKeys(NamedTuple):
    image_version: str
    mgmt_acl_name: str
    mgmt_acl_seq: str
    sla_grp: str
    sla_dst: str
    sla_rtt: str
    sla_state: str


def _set_keys(os_type: str) -> OsKeys:
    """Based on the OS type set the set the dictionary keys when gleaning data from NTC data structures.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
    Returns:
        OsKeys: Dictionary Keys for the specific OS type to retrieve the output data
    """
    if "ios" in os_type:
        return OsKeys(
            "version", "acl_name", "line_num", "interface", "name", "rtt", "status"
        )
    elif "nxos" in os_type:
        return OsKeys("os", "name", "sn", "", "", "", "")
    elif "asa" in os_type:
        return OsKeys("version", "name", "sn", "", "", "", "")
    elif "wlc" in os_type:
        return OsKeys("product_version", "", "", "", "", "", "")
    elif "panos" in os_type:
        return OsKeys(
            "os", "", "", "grp_name", "destination", "rtt_avg", "success_rate"
        )
    # Fallback if nothing matched
    msg = f"Error, '_set_keys' has no match for OS type: '{os_type}'"
    raise NotImplementedError(msg)


# ----------------------------------------------------------------------------
# OUTPUT: Creates str or ntc output dictionaries based on device command output
# ----------------------------------------------------------------------------
def _format_output(
    os_type: str, sub_feature: str, output: list[str | dict[str, str]]
) -> tuple[list[str], list[dict[str, str]]]:
    """Screen scraping return different data structures, they need defining to make function typing easier.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str from raw) command output from the device
    Raises:
        ValueError: Errors if the new output variable doesn't match the input as not meant to be changing it, just defining it for MYPY
    Returns:
        tuple[list[str], list[dict[str, str]]]: Returns either RAW list[str] or NTC list[dict], the other non-matched one will be empty
    """
    raw_output = [o for o in output if isinstance(o, str)]
    ntc_output = [o for o in output if isinstance(o, dict)]
    if (
        len(raw_output) != 0
        and output != raw_output
        or len(ntc_output) != 0
        and output != ntc_output
    ):
        msg = f"{os_type} {sub_feature} _format_output is malformed"
        raise ValueError(msg)
    return raw_output, ntc_output


# ----------------------------------------------------------------------------
# DEF: Mini-functions used by the main function
# ----------------------------------------------------------------------------
def _make_int(input_data: str) -> int | str:
    """Takes a string and returns an integer if it can, otherwise it returns the original string.

    Args:
        input_data (str): The data to be converted to an integer
    Returns:
        int | str: The input_data as a integer if possible, if not as the original string
    """
    try:
        return int(input_data)
    except ValueError:
        return input_data


def _acl_format_into_dict(
    output: list[dict[str, str]], name: str
) -> dict[str, list[dict[str, str]]]:
    """From a list of ACEs, removes the empty dicts and creates a dict of ACLs.

    Args:
        output (list[dict[str, str]]): List of ACE entries
        name (str): Name used in data structure as ACL name for this type of OS
    Returns:
        dict[str, list[dict[str, str]]]: A dict of ACLs with the name of the ACL as the key and a list of ACEs as the value
    """
    acl = {}
    acl_names = set(item[name] for item in output)
    for each_acl_name in acl_names:
        tmp_acl_list = []
        for each_ace in output:
            if each_ace.get("line_num", "dummy") == "":
                pass
            else:
                tmp_ace_dict = {}
                for ace_key, ace_val in each_ace.items():
                    # Add none blank ACE entries to dict of ACEs
                    if len(ace_val) != 0:
                        tmp_ace_dict[ace_key] = ace_val
                # Add dict of all none blank ACE entries to list of ACLs
                if each_acl_name == each_ace[name]:
                    tmp_acl_list.append(tmp_ace_dict)
            # Add all ACLs to dict of ACL names
            acl[each_acl_name] = tmp_acl_list
    return acl


def _acl_remove_remark(
    acl: dict[str, list[dict[str, str]]],
) -> dict[str, list[dict[str, str]]]:
    """Removes remark and reorders the seq number as NXOS gives remarks a sequence number (IOS doesnt).

    Args:
        acl (list[dict[str, str]]): Dictionary if ACLs
    Returns:
        dict[str, list[dict[str, str]]]: Same ACL dict but with remarks removed
    """
    for ace in acl.values():
        while "remark" in str(ace):
            ace_count = len(ace)
            for idx, each_ace in enumerate(ace):
                if each_ace["action"] == "remark":
                    for x in reversed(range(idx + 1, ace_count)):
                        ace[x]["sn"] = ace[x - 1]["sn"]
                    del ace[idx]
                    break
    return acl


def _acl_scr_dst(each_ace: dict[str, str], src_dst: str) -> str:
    """Converts the source or destination address of the ACE converted into address/prefix (IOS).

    Args:
        each_ace (dict[str, str]): Single ACL ACE
        src_dst (str): Either is the source (src) or (destination (dst) address
    Returns:
        str: Returns either the host or network address with prefix rather than subnet mask (or "any" as catchall)
    """
    if each_ace.get(src_dst + "_network") is not None:
        addr = each_ace[src_dst + "_network"] + "/" + each_ace[src_dst + "_wildcard"]
        return ipaddress.IPv4Interface(addr).with_prefixlen
    elif each_ace.get(src_dst + "_host") is not None:
        return ipaddress.IPv4Interface(each_ace[src_dst + "_host"]).with_prefixlen
    else:
        return each_ace[src_dst + "_any"]


def _acl_asa_format(output: list[str], name: str) -> list[dict[str, str]]:
    """Formats ASA SSH and HTTP ACLs into same data structure as normal switch ACLs.

    Args:
        output (list[str]): The command output from the device (raw)
        name (str): Name of access-list being standardised, ssh or http
    Returns:
        list[dict[str, str]]: List of dictionaries with the name, sequence number, action, and source for each ACE (same format as ACLs)
    """
    asa_acl_list = []
    seq = 10
    for each_ace in output:
        if re.match(f"^{name} [0-9]", each_ace):
            asa_dict = dict(name=name)
            asa_dict["sn"] = str(seq)
            asa_dict["action"] = "permit"
            ip_mask = each_ace.split()[1] + "/" + each_ace.split()[2]
            ip_pfxlen = ipaddress.IPv4Interface(ip_mask).with_prefixlen
            if ip_pfxlen == "0.0.0.0/0":
                asa_dict["source"] = f"{each_ace.split()[3]} - any"
            else:
                asa_dict["source"] = f"{each_ace.split()[3]} - {ip_pfxlen}"
            asa_acl_list.append(asa_dict)
            seq = seq + 10
    return asa_acl_list


def _acl_val_file(ace: list[dict[str, str]]) -> dict[str, list[dict[str, str]]]:
    """Creates a list of ace statements from an ACL for the validation file.

    Args:
        ace (list[dict[str, str]]): List of all ACEs for a ACL
    Returns:
        dict[str, list[dict[str, str]]]: Cleaned up ACL ACEs in format [{action: address/pfx}]
    """
    val_aces = []
    for each_ace in ace:
        try:
            val_aces.append({each_ace["action"]: each_ace["source"]})
        # For IOS as needs to convert subnet mask so in the format address/prefix
        except KeyError as e:
            val_aces.append({each_ace["action"]: _acl_scr_dst(each_ace, "src")})
    return dict(ace=val_aces)


def _acl_actual_state(
    ace: list[dict[str, str]], mgmt_acl_seq: str
) -> dict[int | str, dict[str, str]]:
    """Creates a dict of ace statements from an ACL for the actual state grouped by seq (key).

    Args:
        ace (list[dict[str, str]]): List of all ACEs for a ACL
        mgmt_acl_seq (str): os-specific name used to call the ACL
    Returns:
        dict[int | str, dict[str, str]]: Cleaned up ACL ACEs grouped by seq {seq: {'action': x, 'protocol': y, 'dst': 'any', 'src': address/pfx}}
    """
    ac_aces: defaultdict[int | str, dict[str, str]] = defaultdict(dict)
    for each_ace in ace:
        seq = _make_int(each_ace[mgmt_acl_seq])
        ac_aces[seq]["action"] = each_ace["action"]
        ac_aces[seq]["protocol"] = "ip"
        ac_aces[seq]["dst"] = "any"
        try:
            ac_aces[seq]["src"] = each_ace["source"]
        # For IOS as needs to convert subnet mask so in the format address/prefix
        except KeyError as e:
            ac_aces[seq]["src"] = _acl_scr_dst(each_ace, "src")
    return dict(ac_aces)


def _norm_ios_trk_grp(
    key: OsKeys, output: list[dict[str, str]]
) -> list[dict[str, str]]:
    """Split tracker-group objects probes (trackers) into separate dictionaries.

    Args:
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        output (list[dict[str, str]]): The command output from the device in ntc data structure OR raw data structure
    Returns:
        list[dict[str, str]]: List of endpoint trackers grouped by tracker group
    """
    norm_output = []
    for probe in output:
        if probe.get(key.sla_grp) is None:
            # Split each probe element into lists
            rtts = probe[key.sla_rtt].split(", ")
            trackers = probe["trackers"].split(", ")
            tmp_status = probe["status"].split("(")[1].replace(")", "")
            statuses = (
                tmp_status.split(" OR ")
                if " OR " in tmp_status
                else tmp_status.split(" AND ")
            )
            # Loop through all elements simultaneously
            for rtt, status, tracker in zip(rtts, statuses, trackers, strict=False):
                norm_output.append(
                    {
                        key.sla_grp: probe[key.sla_dst],
                        key.sla_dst: tracker,
                        key.sla_rtt: rtt,
                        key.sla_state: status,
                    }
                )
        else:
            norm_output.append(probe)
    return norm_output


def _palo_sla_state(state: str) -> str:
    """Convert Palo Alto SLA success rate state to Up/Down.

    Args:
        state (str): The SLA state from Palo Alto based on probe success rate
    Returns:
        str: Palo SLA state of "Up" or "Down"
    """
    tmp_state = state.split("/")[0] if len(state.split("/")) > 1 else None
    if state.split("/")[0] == tmp_state:
        return "Up"
    else:
        return "Down"


def format_image(key: OsKeys, output: list[dict[str, str]]) -> str:
    """Format image into the data structure.

    Args:
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        output (list[dict[str, str]]): The command output from the device in ntc data structure OR raw data structure
    Returns:
        str:  {image: code_number}
    """
    return output[0][key.image_version]


def format_acl(
    val_file: bool, key: OsKeys, output: list[dict[str, str]]
) -> dict[str, dict[str, list[dict[str, str]]] | dict[int | str, dict[str, str]]]:
    """Format management ACL into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        output (list[dict[str, str]]): The command output from the device
    Returns:
        dict[str, dict[str, list[dict[str, str]]] | dict[int | str, dict[str, str]]]: {seq: {'action': x, 'protocol': y, 'dst': 'any', 'src': address/pfx}},
        val_file is [{action: address/pfx}]
    """
    result: dict[
        str,
        dict[str, list[dict[str, str]]] | dict[int | str, dict[str, str]],
    ] = {}
    # Cleanup up data structure
    acl = _acl_format_into_dict(output, key.mgmt_acl_name)
    acl = _acl_remove_remark(acl)
    # Separate functions to create val file actual state
    for name, ace in acl.items():
        # If creating validation file
        if val_file:
            result[name] = _acl_val_file(ace)
        #  If creating the actual state
        elif not val_file:
            result[name] = _acl_actual_state(ace, key.mgmt_acl_seq)
    return result


def format_module(val_file: bool, output: list[dict[str, str]]) -> dict[str | int, Any]:
    """Format module into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        output (list[dict[str, str]]): The command output from the device in ntc data structure
    Returns:
        dict[str | int, Any]:  {module_num: {model: xxx, status, ok}}, val_file is {module_num: {model: xxx}}
    """
    result: dict[int | str, dict[str, str]] = defaultdict(dict)
    for each_mod in output:
        mod = _make_int(each_mod["module"])
        result[mod]["model"] = each_mod["model"]
        # If creating validation file
        if val_file:
            status = each_mod.get("status", "x").lower()
            if "active" in status or "standby" in status:
                result[mod]["status"] = each_mod["status"].lower()
        #  If creating the actual state
        elif not val_file:
            if len(each_mod["status"]) == 0:
                result[mod]["status"] = "ok"
            else:
                result[mod]["status"] = each_mod["status"].lower()
    return dict(result)


def format_sla(
    val_file: bool, os_type: str, key: OsKeys, output: list[dict[str, str]]
) -> dict[str, Any]:
    """Format SLA monitoring into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        output (list[dict[str, str]]): The command output from the device in ntc data structure OR raw data structure
    Returns:
        dict[str, Any]:  {cmd: {grp_name: {dst: {rtt: x, State: Up}}}}, val_file is {cmd: {grp_name: {dst: {rtt: x}}}}
    """
    result: dict[str, dict[str, dict[str, str | int]]] = defaultdict(dict)
    # Normalise IOS tracker group
    if "ios" in os_type:
        output = _norm_ios_trk_grp(key, output)
    # Run function on all SLAs
    for probe in output:
        try:
            # Round up/down floats as needs to be integer to all for =>, =<, etc
            rtt_num = float(probe[key.sla_rtt])
            result[probe[key.sla_grp]][probe[key.sla_dst]] = {"rtt": round(rtt_num)}
        except (ValueError, TypeError):
            result[probe[key.sla_grp]][probe[key.sla_dst]] = {
                "rtt": _make_int(probe[key.sla_rtt])
            }
        # Add state if it is the actual state
        if not val_file:
            if "panos" in os_type:
                result[probe[key.sla_grp]][probe[key.sla_dst]]["state"] = (
                    _palo_sla_state(probe[key.sla_state])
                )
            else:
                result[probe[key.sla_grp]][probe[key.sla_dst]]["state"] = probe[
                    key.sla_state
                ].capitalize()
    return dict(result)


# ----------------------------------------------------------------------------
# ACTUAL_STATE: Engine use to create sub-feature actual state or validation file
# ----------------------------------------------------------------------------
def format_actual_state(
    val_file: bool,
    os_type: str,
    sub_feature: str,
    output: list[str | dict[str, str]],
) -> str | dict[Any, Any]:
    """Engine to run all the actual state and validation file sub-feature formatting.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str/int from raw) command output from the device
    Returns:
        str | dict[Any, Any]: Returns cmd output formatted into the data structure of actual state or validation file
    """
    key = _set_keys(os_type)
    raw_output, ntc_output = _format_output(os_type, sub_feature, output)

    ### IMAGE: {image: code_number}
    if sub_feature == "image":
        return format_image(key, ntc_output)

    ### MGMT_ACL: {acl_name: seq_num: {protocol: ip/tcp/udp, src: src_ip (or as intf - src_ip), dst: dst_ip, dst_port: port}}
    elif sub_feature == "mgmt_acl":
        if bool(re.search("asa", os_type)):
            ntc_output = _acl_asa_format(raw_output, "ssh")
            ntc_output.extend(_acl_asa_format(raw_output, "http"))
        return format_acl(val_file, key, ntc_output)

    ### MODULE: {module_num: {model: xxx, status, ok}}
    elif sub_feature == "module":
        return format_module(val_file, ntc_output)

    ### SLA: {cmd: {grp_name: {dst: {rtt: x, State: Up}}}}
    if sub_feature == "sla":
        return format_sla(val_file, os_type, key, ntc_output)

    ### CatchAll
    else:
        msg = f"Unsupported sub_feature: {sub_feature}"
        raise ValueError(msg)



from collections import defaultdict
from typing import Any


# ----------------------------------------------------------------------------
# OUTPUT: Creates str or ntc output dictionaries based on device command output
# ----------------------------------------------------------------------------
def _format_output(
    os_type: str, sub_feature: str, output: list[str | dict[str, str]]
) -> tuple[list[str], list[dict[str, str]]]:
    """Screen scraping return different data structures, they need defining to make function typing easier.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str from raw) command output from the device
    Raises:
        ValueError: Errors if the new output variable doesn't match the input as not meant to be changing it, just defining it for MYPY
    Returns:
        tuple[list[str], list[dict[str, str]]]: Returns either RAW list[str] or NTC list[dict], the other non-matched one will be empty
    """
    raw_output = [o for o in output if isinstance(o, str)]
    ntc_output = [o for o in output if isinstance(o, dict)]
    if (
        len(raw_output) != 0
        and output != raw_output
        or len(ntc_output) != 0
        and output != ntc_output
    ):
        msg = f"{os_type} {sub_feature} _format_output is malformed"
        raise ValueError(msg)
    return raw_output, ntc_output


# ----------------------------------------------------------------------------
# DEF: Mini-functions used by the main function
# ----------------------------------------------------------------------------
def _make_int(input_data: str) -> int | str:
    """Takes a string and returns an integer if it can, otherwise it returns the original string.

    Args:
        input_data (str): The data to be converted to an integer
    Returns:
        int | str: The input_data as a integer if possible, if not as the original string
    """
    try:
        return int(input_data)
    except ValueError:
        return input_data


def format_wlan(val_file: bool, output: list[dict[str, str]]) -> dict[str | int, Any]:
    """Format WLANs into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        output (list[dict[str, str]]): The command output from the device in ntc data structure
    Returns:
        dict[str | int, Any]: {wlan_id: {intf: x, ssid: xy, status: Enabled}}, if val_file is {wlan_id: {interface: x, ssid: xy}}
    """
    result: dict[str | int, dict[str, str]] = defaultdict(dict)
    for each_wlan in output:
        wlanid = _make_int(each_wlan["wlanid"])
        result[wlanid]["ssid"] = each_wlan["ssid"]
        # If creating actual state (replace(none) needed as long names merge next column (PIMIPv6))
        if not val_file:
            result[wlanid]["intf"] = each_wlan["interface"].replace("none", "").rstrip()
            result[wlanid]["status"] = "Enabled"
        # If creating validation file
        elif val_file:
            result[wlanid]["interface"] = (
                each_wlan["interface"].replace("none", "").rstrip()
            )
    return dict(result)


def format_ap(output: list[dict[str, str]]) -> dict[str, Any]:
    """Format APs into the data structure.

    Args:
        output (list[dict[str, str]]): The command output from the device in ntc data structure
    Returns:
        dict[str, Any]: {ap_name: {model: x, ip: x, clients: x}}}
    """
    result: dict[str, dict[str, str | int]] = defaultdict(dict)
    for each_ap in output:
        result[each_ap["ap_name"]]["model"] = each_ap["ap_model"]
        result[each_ap["ap_name"]]["ip"] = each_ap["ip_address"]
        result[each_ap["ap_name"]]["client_count"] = _make_int(each_ap["clients"])
    return dict(result)


def format_client_count(output: list[str]) -> dict[str, str | int]:
    """Format client count into the data structure.

    Args:
        output (list[str]): The command output from the device in raw data structure
    Returns:
        dict[str, str | int]: {total_count: x, wlxx_count: x}
    """
    result = {}
    tmp_output = []
    # Gathers AP count from raw screen scrapped output
    for each_line in output:
        if "Number of Cl" in each_line or "Invalid WLAN ID 999" in each_line:
            tmp_output.append(each_line)
    # Formats client count from the slimmed down data
    if len(tmp_output) != 0:
        for idx, each_line in enumerate(tmp_output):
            if "Invalid WLAN ID 999" in each_line:
                name = f"wl{each_line.split()[4].replace('999', '')}_count"
                result[name] = _make_int(tmp_output[idx + 1].split()[-1])
                tmp_output[idx + 1] = ""
            elif "Number of Cl" in each_line:
                result["total_count"] = _make_int(each_line.split()[-1])
    return result


def format_flexconnect(output: list[dict[str, str]]) -> dict[str, Any]:
    """Format flexconnect groups into the data structure.

    Args:
        output (list[dict[str, str]]): The command output from the device in ntc data structure
    Returns:
        dict[str, Any]: {grp_name: {ap_count: x}}}
    """
    result: dict[str, dict[str, str | int]] = defaultdict(dict)
    for each_grp in output:
        name = each_grp["flexconnect_group_name"]
        result[name]["ap_count"] = _make_int(each_grp["ap_count"])
    return dict(result)


def format_intf_grp(output: list[dict[str, str]]) -> dict[str, Any]:
    """Format interface groups into the data structure.

    Args:
        output (list[dict[str, str]]): The command output from the device in ntc data structure
    Returns:
        dict[str, Any]: {ap_count: x, intf_count: x, wlan_count: x}}}
    """
    result: dict[str, dict[str, str | int]] = defaultdict(dict)
    for each_grp in output:
        name = each_grp["interface_group_name"]
        result[name]["ap_grp_count"] = _make_int(each_grp["total_ap_groups"])
        result[name]["intf_count"] = _make_int(each_grp["total_interfaces"])
        result[name]["wlan_count"] = _make_int(each_grp["total_wlans"])
    return dict(result)


# ----------------------------------------------------------------------------
# ACTUAL_STATE: Engine use to create sub-feature actual state or validation file
# ----------------------------------------------------------------------------
def format_actual_state(
    val_file: bool,
    os_type: str,  # noqa: ARG001
    sub_feature: str,
    output: list[str | dict[str, str]],
) -> dict[Any, Any]:
    """Engine to run all the actual state and validation file sub-feature formatting.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str/int from raw) command output from the device
    Returns:
        dict[str, Any]: Returns cmd output formatted into the data structure of actual state or validation file
    """
    raw_output, ntc_output = _format_output(os_type, sub_feature, output)
    ### WLAN: {wlan_id: {intf: x, ssid: xy, status: Enabled}}}, if val_file is {wlan_id: {interface: x, ssid: xy}}
    if sub_feature == "wlan":
        return format_wlan(val_file, ntc_output)

    ### AP: {ap_name: {model: x, ip: x, clients: x}}}
    elif sub_feature == "ap":
        return format_ap(ntc_output)

    ### CLIENT_COUNT: {total_count: x, wlxx_count: x}
    elif sub_feature == "client_count":
        return format_client_count(raw_output)

    ### FLEXCONN: {grp_name: {ap_count: x}}}
    elif sub_feature == "flexconnect":
        return format_flexconnect(ntc_output)

    ### INTF_GRP: {grp_name: {ap_count: x, intf_count: x, wlan_count: x}}}
    elif sub_feature == "intf_grp":
        return format_intf_grp(ntc_output)

    ### CatchAll
    else:
        msg = f"Unsupported sub_feature: {sub_feature}"
        raise ValueError(msg)



from collections import defaultdict
from typing import Any, NamedTuple


# ----------------------------------------------------------------------------
# KEY: Set dictionary keys on a per-os_type basis
# ----------------------------------------------------------------------------
class OsKeys(NamedTuple):
    cdp_local: str
    cdp_nbr: str
    cdp_remote: str


def _set_keys(os_type: str) -> OsKeys:
    """Based on the OS type set the set the dictionary keys when gleaning data from NTC data structures.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
    Returns:
        OsKeys: Dictionary Keys for the specific OS type to retrieve the output data
    """
    if "ios" in os_type or "nxos" in os_type:
        return OsKeys("local_interface", "neighbor_name", "neighbor_interface")
    elif "wlc" in os_type:
        return OsKeys("local_interface", "chassis_id", "neighbor_interface")
    # Fallback if nothing matched
    msg = f"Error, '_set_keys' has no match for OS type: '{os_type}'"
    raise NotImplementedError(msg)


# ----------------------------------------------------------------------------
# OUTPUT: Creates str or ntc output dictionaries based on device command output
# ----------------------------------------------------------------------------
def _format_output(
    os_type: str, sub_feature: str, output: list[str | dict[str, str]]
) -> tuple[list[str], list[dict[str, str]]]:
    """Screen scraping return different data structures, they need defining to make function typing easier.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str from raw) command output from the device
    Raises:
        ValueError: Errors if the new output variable doesn't match the input as not meant to be changing it, just defining it for MYPY
    Returns:
        tuple[list[str], list[dict[str, str]]]: Returns either RAW list[str] or NTC list[dict], the other non-matched one will be empty
    """
    raw_output = [o for o in output if isinstance(o, str)]
    ntc_output = [o for o in output if isinstance(o, dict)]
    if (
        len(raw_output) != 0
        and output != raw_output
        or len(ntc_output) != 0
        and output != ntc_output
    ):
        msg = f"{os_type} {sub_feature} _format_output is malformed"
        raise ValueError(msg)
    return raw_output, ntc_output


# ----------------------------------------------------------------------------
# DEF: Mini-functions used by the main function
# ----------------------------------------------------------------------------
def format_cdp_lldp_nbr(key: OsKeys, output: list[dict[str, str]]) -> dict[str, Any]:
    """Format CDP or LLDP neighbors into the data structure.

    Args:
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        output (list[dict[str, Any]]): The command output from the device in ntc data structure
    Returns:
        dict[str, Any]: {intf: {nbr_name: xx, nbr_intf: yy}}
    """
    result: dict[str, dict[str, str]] = defaultdict(dict)
    for each_nbr in output:
        result[each_nbr[key.cdp_local]]["nbr_name"] = each_nbr[key.cdp_nbr]
        result[each_nbr[key.cdp_local]]["nbr_intf"] = each_nbr[key.cdp_remote]
    return dict(result)


# ----------------------------------------------------------------------------
# ACTUAL_STATE: Engine use to create sub-feature actual state or validation file
# ----------------------------------------------------------------------------
def format_actual_state(
    val_file: bool,  # noqa: ARG001
    os_type: str,
    sub_feature: str,
    output: list[str | dict[str, str]],
) -> dict[str, Any]:
    """Engine to run all the actual state and validation file sub-feature formatting.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str/int from raw) command output from the device
    Returns:
        dict[str, dict[str, str]]: Returns cmd output formatted into the data structure of actual state or validation file
    """
    key = _set_keys(os_type)
    raw_output, ntc_output = _format_output(os_type, sub_feature, output)

    ### CDP/LLDP: {intf: {nbr_name: xx, nbr_intf: yy}}
    if sub_feature == "cdp" or sub_feature == "lldp":
        return format_cdp_lldp_nbr(key, ntc_output)
    ### CatchAll
    else:
        msg = f"Unsupported sub_feature: {sub_feature}"
        raise ValueError(msg)



from collections import defaultdict
from typing import Any, NamedTuple


# ----------------------------------------------------------------------------
# KEY: Set dictionary keys on a per-os_type basis
# ----------------------------------------------------------------------------
class OsKeys(NamedTuple):
    omp_peer: str
    cntl_nhbr: str
    cntl_color: str
    bfd_nhbr: str


def _set_keys(os_type: str) -> OsKeys:
    """Based on the OS type set the set the dictionary keys when gleaning data from NTC data structures.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
    Returns:
        OsKeys: Dictionary Keys for the specific OS type to retrieve the output data
    """
    if "ios" in os_type:  # noqa: SIM114
        return OsKeys("peer", "system_ip", "local_color", "system_ip")
    elif "viptela" in os_type:
        return OsKeys("peer", "system_ip", "remote_color", "system_ip")
    # Fallback if nothing matched
    msg = f"Error, '_set_keys' has no match for OS type: '{os_type}'"
    raise NotImplementedError(msg)


# ----------------------------------------------------------------------------
# OUTPUT: Creates str or ntc output dictionaries based on device command output
# ----------------------------------------------------------------------------
def _format_output(
    os_type: str, sub_feature: str, output: list[str | dict[str, str]]
) -> tuple[list[str], list[dict[str, str]]]:
    """Screen scraping return different data structures, they need defining to make function typing easier.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str from raw) command output from the device
    Raises:
        ValueError: Errors if the new output variable doesn't match the input as not meant to be changing it, just defining it for MYPY
    Returns:
        tuple[list[str], list[dict[str, str]]]: Returns either RAW list[str] or NTC list[dict], the other non-matched one will be empty
    """
    raw_output = [o for o in output if isinstance(o, str)]
    ntc_output = [o for o in output if isinstance(o, dict)]
    if (
        len(raw_output) != 0
        and output != raw_output
        or len(ntc_output) != 0
        and output != ntc_output
    ):
        msg = f"{os_type} {sub_feature} _format_output is malformed"
        raise ValueError(msg)
    return raw_output, ntc_output


# ----------------------------------------------------------------------------
# DEF: Mini-functions used by the main function
# ----------------------------------------------------------------------------
def _make_int(input_data: str) -> int | str:
    """Takes a string and returns an integer if it can, otherwise it returns the original string.

    Args:
        input_data (str): The data to be converted to an integer
    Returns:
        int | str: The input_data as a integer if possible, if not as the original string
    """
    try:
        return int(input_data)
    except ValueError:
        return input_data


def format_control_conn(
    val_file: bool, key: OsKeys, output: list[dict[str, str]]
) -> dict[str, Any]:
    """Format control connections into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        output (list[dict[str, str]]): The command output from the device in ntc data structure OR raw data structure
    Returns:
        dict[str, dict[str, str | int]]: {nhbr: {site_id:x, color:x, state:x}} val file has no {state: x}
    """
    result: dict[str, dict[str, str | int]] = defaultdict(dict)
    for entry in output:
        nhbr = entry[key.cntl_nhbr]
        result[nhbr]["site_id"] = _make_int(entry["site_id"])
        result[nhbr]["color"] = entry[key.cntl_color]
        # If is actual_state adds neighbor state
        if not val_file:
            result[nhbr]["state"] = entry["state"]
    return dict(result)


def format_omp(
    val_file: bool, key: OsKeys, output: list[dict[str, str]]
) -> dict[str, Any]:
    """Format OMP peer into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        output (list[dict[str, str]]): The command output from the device in ntc data structure OR raw data structure
    Returns:
        dict[str, dict[str, str | int]]: {peer: {site_id:x, routes_received:x, routes_installed:x:, routes_sent:x, state:x}} val file has no {state: x}
    """
    result: dict[str, dict[str, str | int]] = defaultdict(dict)
    for entry in output:
        peer = entry[key.omp_peer]
        result[peer]["site_id"] = _make_int(entry["site_id"])
        result[peer]["routes_received"] = _make_int(entry["routes_received"])
        result[peer]["routes_installed"] = _make_int(entry["routes_installed"])
        result[peer]["routes_sent"] = _make_int(entry["routes_sent"])
        # If is actual_state adds peer state
        if not val_file:
            result[peer]["state"] = entry["state"]
    return dict(result)


def format_bfd_session(
    val_file: bool, key: OsKeys, output: list[dict[str, str]]
) -> dict[str, Any]:
    """Format BFD sessions into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        output (list[dict[str, str]]): The command output from the device in ntc data structure OR raw data structure
    Returns:
        dict[str, dict[str, str | int]]: {nhbr: {site_id:x, local_color:x, remote_color:x, state:x}} val file has no {state: x}
    """
    result: dict[str, dict[str, str | int]] = defaultdict(dict)
    for entry in output:
        nhbr = entry[key.bfd_nhbr]
        result[nhbr]["site_id"] = _make_int(entry["site_id"])
        result[nhbr]["local_color"] = entry["local_color"]
        result[nhbr]["remote_color"] = entry["remote_color"]
        # If is actual_state adds neighbor state
        if not val_file:
            result[nhbr]["state"] = entry["state"]
    return dict(result)


# ----------------------------------------------------------------------------
# ACTUAL_STATE: Engine use to create sub-feature actual state or validation file
# ----------------------------------------------------------------------------
def format_actual_state(
    val_file: bool,
    os_type: str,
    sub_feature: str,
    output: list[str | dict[str, str]],
) -> dict[str, Any]:
    """Engine to run all the actual state and validation file sub-feature formatting.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str/int from raw) command output from the device
    Returns:
        dict[str, Any]: Returns cmd output formatted into the data structure of actual state or validation file
    """
    key = _set_keys(os_type)
    raw_output, ntc_output = _format_output(os_type, sub_feature, output)

    ### CONTROL_CONN: {nhbr: {site_id:x, color:x, state:x}}
    if sub_feature == "control_conn":
        return format_control_conn(val_file, key, ntc_output)
    ### OMP_PEER: {peer: {site_id:x, routes_received:x, routes_installed:x:, routes_sent:x, state:x}}
    elif sub_feature == "omp_peer":
        return format_omp(val_file, key, ntc_output)
    ### BFD_SESSION: {nhbr: {site_id:x, local_color:x, remote_color:x, state:x}}
    elif sub_feature == "bfd_session":
        return format_bfd_session(val_file, key, ntc_output)
    ### CatchAll
    else:
        msg = f"Unsupported sub_feature: {sub_feature}"
        raise ValueError(msg)



import re
from collections import defaultdict
from typing import Any


# ----------------------------------------------------------------------------
# DEF: Mini-functions used by the main function
# ----------------------------------------------------------------------------
def _make_int(input_data: str) -> int | str:
    """Takes a string and returns an integer if it can, otherwise it returns the original string.

    Args:
        input_data (str): The data to be converted to an integer
    Returns:
        int | str: The input_data as a integer if possible, if not as the original string
    """
    try:
        return int(input_data)
    except ValueError:
        return input_data


def format_sts_peer(
    val_file: bool,
    os_type: str,
    output: list[dict[str, Any]],
) -> dict[str, str] | list[str]:
    """Format STS peers into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        output (list[dict[str, Any]]): The command output from the device in ntc data structure
    Returns:
        dict[str, str] | list[str]: {peer_ip: IP, peer_ip: UP-ACTIVE}}, val file is just list a [peer_ip, peer_ip]
    """
    result = {}
    # For ASA need to standardise the session status based on receive/transmit packets
    if re.search("asa", os_type):
        sts_output = []
        for each_peer in output:
            if (
                len(each_peer["udp_src_port"]) != 0
                and len(each_peer["udp_dst_port"]) != 0
            ):
                tmp_dict = {
                    "peer": each_peer["connection"],
                    "session_status": "UP-ACTIVE",
                }
                if (
                    each_peer["total_bytes_received"] == "0"
                    and each_peer["total_bytes_transmitted"] == "0"
                ):
                    tmp_dict["session_status"] = "DOWN-NoRX/TX"
                elif each_peer["total_bytes_received"] == "0":
                    tmp_dict["session_status"] = "DOWN-NoRX"
                elif each_peer["total_bytes_transmitted"] == "0":
                    tmp_dict["session_status"] = "DOWN-NoTX"
                sts_output.append(tmp_dict)
    # All other OS types change output name so can all use same loop
    else:
        sts_output = output
    val_file_list = []
    for each_vpn in sts_output:
        result[each_vpn["peer"]] = each_vpn["session_status"]
        val_file_list.append(each_vpn["peer"])
    # List of peers if is a validation file
    if val_file:
        return val_file_list
    else:
        return result


def format_ac_user(output: list[dict[str, Any]]) -> dict[str, Any]:
    """Format AnyConnect Users into the data structure.

    Args:
        output (list[dict[str, Any]]): The command output from the device in ntc data structure
    Returns:
        dict[str, Any]: {user: tunnel_group: x, group_policy:y }}
    """
    result: dict[str, dict[str, str]] = defaultdict(dict)
    for each_vpn in output:
        user = each_vpn["username"]
        result[user]["group_policy"] = each_vpn["group_policy"]
        result[user]["tunnel_group"] = each_vpn["tunnel_group"]
    return dict(result)


def format_vpn_count(
    output: list[str | dict[str, Any]],
) -> dict[str, str | int]:
    """Format VPN count for AC and StS into the data structure.

    Args:
        output (list[str | dict[str, Any]]): The command output from the device, 2 commands so includes raw str and ntc dict
    Returns:
        dict[str, str | int]: {sts: x, ac:y }}
    """
    result = {}
    for each_line in output:
        if isinstance(each_line, str):
            if "Number of lines which match regexp" in each_line:
                result["sts"] = _make_int(each_line.split()[-1])
        else:
            for each_vpn, active_sess in zip(
                each_line["vpn_session_name"],
                each_line["vpn_session_active"],
                strict=False,
            ):
                if each_vpn == "AnyConnect Client":
                    result["ac"] = _make_int(active_sess)
    return result


# ----------------------------------------------------------------------------
# ACTUAL_STATE: Engine use to create sub-feature actual state or validation file
# ----------------------------------------------------------------------------
def format_actual_state(
    val_file: bool, os_type: str, sub_feature: str, output: list[dict[str, Any]]
) -> dict[str, Any] | list[str]:
    """Engine to run all the actual state and validation file sub-feature formatting.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[dict[str, Any]]): The structured (dict from NTC template) or unstructured (str/int from raw) command output from the device
    Returns:
        dict[str, Any] | list[str]: Returns formatted command output based on sub_feature and val_file arguments.
    """
    ### STS_PEER: {peer_ip: IP, peer_ip: UP-ACTIVE}} (val file is just list a [peer_ip, peer_ip]).
    if sub_feature == "sts_peer":
        return format_sts_peer(val_file, os_type, output)

    ### AC_USER: {user: tunnel_group: x, group_policy:y }}
    elif sub_feature == "ac_user":
        return format_ac_user(output)

    ### VPN_COUNT: {sts: x, ac:y }}
    elif sub_feature == "vpn_count":
        # Reforming output so it has the right typing
        mixed_output: list[str | dict[str, Any]] = [
            o for o in output if isinstance(o, (str, dict))
        ]
        return format_vpn_count(mixed_output)

    ### CatchAll
    else:
        msg = f"Unsupported sub_feature: {sub_feature}"
        raise ValueError(msg)



import json
import os
import re
from datetime import datetime
from typing import Any

from napalm.base import validate  # type: ignore


# ----------------------------------------------------------------------------
# FIX: napalm_validate doesn't recognize ~/ for home drive
# ----------------------------------------------------------------------------
def fix_home_path(input_path: str) -> str:
    """If the string starts with `~/`, it replaces the `~` with the user's home directory (due to napalm bug #1180).

    Arg:
        input_path (str): The path to the input file

    Return (str): The input path with the home directory expanded
    """
    if re.match("^~/", input_path):
        return os.path.expanduser(input_path)
    else:
        return input_path


# -------------------------------------------------------------------------------------------
# REPORT_FILE: If a hostname and directory are passed in as function arguments saves report to file
# --------------------------------------------------------------------------------------------
def save_report_to_file(
    hostname: str,
    directory: str,
    report: dict[str, Any],
    complies: bool,
    skipped: list[str],
) -> str:
    """If report file exists updates the 'skipped' and 'complies' keys, else creates a report with 'skipped' and 'complies' keys.

    Args:
        hostname (str): The hostname of the device running the validation against
        directory (str): The directory where the report will be saved
        report (dict[str, Any]): The report dictionary that is returned by napalm_validate to be used by nornir Result
        complies (bool): True if the device complies with the validation rules, False otherwise
        skipped (list[str]): A list of skipped tests

    Return: Report location and command to view prettified version (xxx | python -m json.tool)
    """
    filename = os.path.join(
        fix_home_path(directory),
        hostname
        + "_compliance_report_"
        + datetime.now().strftime("%Y%m%d-%H%M")
        + ".json",
    )
    # If report file already exists conditionally updates 'skipped' and 'complies' with report outcome
    if os.path.exists(filename):
        with open(filename) as file_content:
            existing_report = json.load(file_content)
        if list(report.values())[0].get("skipped"):
            existing_report["skipped"].extend(skipped)
        elif len(skipped) == 0:
            existing_report["skipped"] = skipped
        # Only adds if is no already failing compliance
        if existing_report.get("complies"):
            existing_report["complies"] = complies
    # If creating a new report file adds 'complies' and 'skipped' with report outcome
    else:
        existing_report = {}
        existing_report["complies"] = complies
        if len(skipped) == 0:
            existing_report["skipped"] = skipped
    # Writes to file the full napalm_validate result (including an existing report)
    existing_report.update(report)
    with open(filename, "w") as file_content:
        json.dump(existing_report, file_content)
    # return f" The report can be viewed using:  \n \33[3m\033[1;37m\33[30m  cat {filename} | python -m json.tool \033[0;0m"
    return f" The report can be viewed using:  \ncat {filename} | python -m json.tool"


# ----------------------------------------------------------------------------------------------------------
# VALIDATE: Uses naplam_validate on custom data fed in (still supports '_mode: strict') to validate and create reports
# ----------------------------------------------------------------------------------------------------------
def generate_validate_report(
    d_state: dict[str, Any],
    a_state: dict[str, Any],
    hostname: str,
    directory: str | None,
) -> dict[str, Any]:
    """Runs the napalm-validate compare method on each feature, adds skipped key if cant be run producing compliance report output based on all features.

    Args:
        d_state (dict[str, Any]): Desired state got from the user input
        a_state (dict[str, Any]): Actual state got from the device
        hostname (str): Hostname of the device being validated
        directory (str | None): If specified the directory where the report will be saved

    Returns (dict[str, Any]): A dictionary of report details result (compliance state) and tasks status, all all fed into Nornir Result
    """
    report: dict[str, Any] = {}
    for feature, sub_feat in d_state.items():
        for each_sub_feat in sub_feat:
            try:
                name = f"{feature}.{each_sub_feat}"
                # napalm_validate compare method produces report based on desired and actual state
                d_state_sub_feat = d_state[feature][each_sub_feat]
                a_state_sub_feat = a_state[feature][each_sub_feat]
                if isinstance(d_state_sub_feat, dict):
                    report[name] = validate.compare(d_state_sub_feat, a_state_sub_feat)
                else:
                    report[name] = validate.compare(
                        {each_sub_feat: d_state_sub_feat},
                        {each_sub_feat: a_state_sub_feat},
                    )
            # If validation couldn't be run on a command adds skipped key to the cmd dictionary
            except NotImplementedError:
                report[feature] = {"skipped": True, "reason": "NotImplemented"}
    # RESULT: Results of compliance report (complies = validation result, skipped (list of skipped cmds) = validation didn't run)
    complies = all([each_cmpl.get("complies", True) for each_cmpl in report.values()])
    skipped = [feat for feat, output in report.items() if output.get("skipped", False)]
    # REPORT_FILE: Save report to file, if not add complies and skipped dictionary to report
    if hostname is not None and directory is not None:
        report_text = save_report_to_file(
            hostname, directory, report, complies, skipped
        )
    else:
        # Empty value if report_file not created
        report_text = ""
    # These must be added after the report
    report["complies"] = complies
    if len(skipped) != 0:
        report["skipped"] = skipped
    # RETURN_RESULT: If compliance fails set state failed (used by Nornir). report dict is used in validation builder
    if complies:
        my_report = dict(
            failed=False,
            result="\u2705 Compliance report complies, desired_state and actual_state match.",
            report=report,
            report_text=report_text,
        )
    if not complies or skipped:
        my_report = dict(
            failed=True, result=report, report=report, report_text=report_text
        )
    return my_report



# ----------------------------------------------------------------------------
# DEF: Mini-functions used by the main function
# ----------------------------------------------------------------------------
def _make_int(input_data: str) -> int | str:
    """Takes a string and returns an integer if it can, otherwise it returns the original string.

    Args:
        input_data (str): The data to be converted to an integer
    Returns:
        int | str: The input_data as a integer if possible, if not as the original string
    """
    try:
        return int(input_data)
    except ValueError:
        return input_data


# ----------------------------------------------------------------------------
# ACTUAL_STATE: Engine use to create sub-feature actual state or validation file
# ----------------------------------------------------------------------------
def format_actual_state(
    val_file: bool,  # noqa: ARG001
    os_type: str,  # noqa: ARG001
    sub_feature: str,
    output: list[str],
) -> int | str:
    """Takes cmd output and just returns the numerical value.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str]): The structured (dict from NTC template) or unstructured (str/int from raw) command output from the device
    Returns:
        int | str: {mab_count: xx}
    """
    ### MAB_COUNT: {mab_count: xx}
    if sub_feature == "mab_count" or sub_feature == "dot1x_count":
        return _make_int(output[0].split()[-1])

    ### CatchAll
    else:
        msg = f"Unsupported sub_feature: {sub_feature}"
        raise ValueError(msg)



import re
from collections import defaultdict
from typing import Any, NamedTuple


# ----------------------------------------------------------------------------
# KEY: Set dictionary keys on a per-os_type basis
# ----------------------------------------------------------------------------
class OsKeys(NamedTuple):
    intf_name: str
    intf_type: str
    intf_status: str
    ip_name: str
    ip_ip: str
    ip_status: str


def _set_keys(os_type: str) -> OsKeys:
    """Based on the OS type set the set the dictionary keys when gleaning data from NTC data structures.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
    Returns:
        OsKeys: Dictionary Keys for the specific OS type to retrieve the output data
    """
    if "ios" in os_type:
        return OsKeys("port", "vlan_id", "status", "interface", "ip_address", "status")
    elif "nxos" in os_type:
        return OsKeys("port", "vlan_id", "status", "intf-name", "prefix", "link-state")
    elif "asa" in os_type:
        return OsKeys(
            "interface",
            "interface_zone",
            "link_status",
            "interface",
            "ip_address",
            "status",
        )
    elif "wlc" in os_type:
        return OsKeys(
            "port", "stp_status", "link_status", "name", "ip_address", "status"
        )
    elif "panos" in os_type:
        return OsKeys("interface", "", "state", "interface", "ip_address", "state")
    # Fallback if nothing matched
    msg = f"Error, '_set_keys' has no match for OS type: '{os_type}'"
    raise NotImplementedError(msg)


# ----------------------------------------------------------------------------
# OUTPUT: Creates str or ntc output dictionaries based on device command output
# ----------------------------------------------------------------------------
def _format_output(
    os_type: str, sub_feature: str, output: list[str | dict[str, str]]
) -> tuple[list[str], list[dict[str, str]]]:
    """Screen scraping return different data structures, they need defining to make function typing easier.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str from raw) command output from the device
    Raises:
        ValueError: Errors if the new output variable doesn't match the input as not meant to be changing it, just defining it for MYPY
    Returns:
        tuple[list[str], list[dict[str, str]]]: Returns either RAW list[str] or NTC list[dict], the other non-matched one will be empty
    """
    raw_output = [o for o in output if isinstance(o, str)]
    ntc_output = [o for o in output if isinstance(o, dict)]
    if (
        len(raw_output) != 0
        and output != raw_output
        or len(ntc_output) != 0
        and output != ntc_output
    ):
        msg = f"{os_type} {sub_feature} _format_output is malformed"
        raise ValueError(msg)
    return raw_output, ntc_output


# ----------------------------------------------------------------------------
# DEF: Mini-functions used by the main function
# ----------------------------------------------------------------------------
def _make_int(input_data: str) -> int | str:
    """Takes a string and returns an integer if it can, otherwise it returns the original string.

    Args:
        input_data (str): The data to be converted to an integer
    Returns:
        int | str: The input_data as a integer if possible, if not as the original string
    """
    try:
        return int(input_data)
    except ValueError:
        return input_data


def _fix_nxos(
    main_dict: dict[str, Any], parent_name: str, child_name: str
) -> list[dict[str, Any]]:
    """NXOS JSON makes dict rather than list if only 1 item in output. If the child_dict is a dict this methods convert it to a list.

    Args:
        main_dict (dict[str, Any]): The dictionary of the command output
        parent_name (str): The parent dictionary key name (e.g. TABLE_xx)
        child_name (str): The key of the dictionary that you want to fix, for example ROW_xx
    Returns:
        A list of dictionaries
    """
    child_dict_value: list[dict[str, Any]] | dict[str, Any] = main_dict[parent_name][
        child_name
    ]
    if isinstance(child_dict_value, dict):
        child_dict_value = [child_dict_value]
    return child_dict_value


def _make_none(
    input_dict: dict[str, str | int | None], dict_key: str
) -> str | int | None:
    """Checks dict key value and if empty returns None.

    Args:
        input_dict(dict[str, str | int | None]): Input dictionary of key-value pairs
        dict_key(key): The dict key that want to check
    Returns:
        str | int | None: The key value or if empty None
    """
    tmp_value = input_dict.get(dict_key, "")
    if tmp_value is None:
        return None
    if isinstance(tmp_value, str):
        return tmp_value if tmp_value else None
    return tmp_value  # keep int values as-is


def format_intf(
    val_file: bool, os_type: str, key: OsKeys, output: list[dict[str, Any]]
) -> dict[str | int, Any]:
    """Format interface output into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        output (list[dict[str, Any]]): The command output from the device in ntc data structure
    Returns:
        dict[str | int, Any]:  {intf: {duplex: x, speed: x, type:x, status: connected }}, val_file doesn't have 'status'
    """
    result: dict[str | int, dict[str, str | int | None]] = defaultdict(dict)
    skip_statuses = {
        "disabled",
        "xcvrAbsen",
        "administratively down",
        "admin down",
        "Down",
        "notconnec",
        "notconnect",
        "down",
    }
    for each_intf in output:
        intf_name = _make_int(each_intf[key.intf_name])
        # BYPASS: Dont include disabled ports when building val_file
        if val_file and each_intf[key.intf_status] in skip_statuses:
            continue
        # SPEED/DUPLEX: WLC doesnt have standard duplex and speed NTC keys
        if bool(re.search("wlc", os_type)):
            result[intf_name]["speed"] = _make_int(
                each_intf["physical_status"].split()[0]
            )
            try:
                duplex = _make_int(each_intf["physical_status"].split()[1])
                result[intf_name]["duplex"] = duplex
            except IndexError:
                result[intf_name]["duplex"] = each_intf["physical_status"]
        # SPEED/DUPLEX: actual file
        elif not val_file:
            result[intf_name]["duplex"] = _make_none(each_intf, "duplex")
            result[intf_name]["speed"] = _make_int(each_intf["speed"])
            result[intf_name]["speed"] = _make_none(result[intf_name], "speed")
        # SPEED/DUPLEX: Val file
        elif val_file:
            if len(each_intf["duplex"]) != 0:
                result[intf_name]["duplex"] = each_intf["duplex"]
            if len(each_intf["speed"]) != 0:
                result[intf_name]["speed"] = _make_int(each_intf["speed"])
        # TYPE: Applies to actual_state and val_file
        if isinstance(_make_int(each_intf.get(key.intf_type, "x")), int):
            result[intf_name]["type"] = "access"
        # TYPE: Actual_state only
        elif not val_file:
            result[intf_name]["type"] = _make_none(each_intf, key.intf_type)
        # TYPE: VAL_file only
        elif len(each_intf.get(key.intf_type, "")) != 0:
            result[intf_name]["type"] = each_intf[key.intf_type]
        # STATUS: Applies all Actual_state
        if not val_file:
            result[intf_name]["status"] = (
                each_intf[key.intf_status].lower().replace("up", "connected")
            )
    return dict(result)


def format_swport(val_file: bool, output: list[dict[str, Any]]) -> dict[str, Any]:
    """Format switchport output into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        output (list[dict[str, Any]]): The command output from the device in ntc data structure
    Returns:
        dict[str, Any]:  {intf: {mode: access or trunk, vlan: x or [x,y]}}
    """
    result: dict[str, dict[str, str | int | list[str] | None]] = defaultdict(dict)

    for each_intf in output:
        # BYPASS: Dont include disabled ports when building val_file
        skip_statuses = {"down", ""}
        if val_file and each_intf["mode"] in skip_statuses:
            continue
        if len(each_intf["mode"]) == 0:
            continue
        else:
            mode = each_intf["mode"].replace("static ", "").split()[0]
        result[each_intf["interface"]]["mode"] = mode
        if each_intf["mode"] == "static access" or each_intf["mode"] == "access":
            result[each_intf["interface"]]["vlan"] = _make_int(each_intf["access_vlan"])
        elif bool(re.search("trunk", each_intf["mode"])):
            trunk_vl = each_intf["trunking_vlans"]
            try:
                trunk_vl = [_make_int(vl) for vl in trunk_vl.split(",")]
            except AttributeError:
                trunk_vl = [_make_int(vl) for vl in trunk_vl[0].split(",")]
            result[each_intf["interface"]]["vlan"] = trunk_vl
        else:
            result[each_intf["interface"]]["vlan"] = None
    return dict(result)


def format_ipbrief(
    val_file: bool, os_type: str, key: OsKeys, output: list[dict[str, Any]]
) -> dict[str, Any]:
    """Format ip interface brief output into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        output (list[dict[str, Any]]): The command output from the device in ntc data structure
    Returns:
        dict[str, Any]: {intf: {ip:x, status: x}}, val_file is {intfx: ipx, intfy: ipy}
    """
    result: dict[str, Any] = defaultdict(dict)
    skip_conditions = {
        key.ip_ip: ["unassigned", "N/A"],
        key.ip_status: ["administratively down", "admin-down"],
    }
    if bool(re.search("nxos", os_type)):
        output = _fix_nxos(output[0], "TABLE_intf", "ROW_intf")
    for each_intf in output:
        # Val file only
        if val_file:
            if any(
                each_intf.get(field) in values
                for field, values in skip_conditions.items()
            ):
                continue
            else:
                # As palo gets state from intf cmd, needed to stop IP from being wiped
                if (
                    bool(re.search("panos", os_type))
                    and each_intf.get(key.ip_ip) is None
                ):
                    pass
                else:
                    result[each_intf[key.ip_name]] = each_intf.get(
                        key.ip_ip, each_intf.get("unnum-intf")
                    )
        # Actual state only
        elif not val_file:
            # As palo gets state from intf cmd, needed to stop IP from being wiped
            if bool(re.search("panos", os_type)) and each_intf.get(key.ip_ip) is None:
                pass
            else:
                result[each_intf[key.ip_name]]["ip"] = each_intf.get(
                    key.ip_ip, each_intf.get("unnum-intf")
                )
            result[each_intf[key.ip_name]]["status"] = each_intf.get(
                key.ip_status, "up"
            )
    return dict(result)


# ----------------------------------------------------------------------------
# ACTUAL_STATE: Engine use to create sub-feature actual state or validation file
# ----------------------------------------------------------------------------
def format_actual_state(
    val_file: bool,
    os_type: str,
    sub_feature: str,
    output: list[str | dict[str, str]],
) -> dict[Any, Any]:
    """Engine to run all the actual state and validation file sub-feature formatting.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str/int from raw) command output from the device
    Returns:
        dict[str | int, Any]: Returns cmd output formatted into the data structure of actual state or validation file
    """
    key = _set_keys(os_type)
    aw_output, ntc_output = _format_output(os_type, sub_feature, output)

    ### INTF: {intf: {duplex: x, speed: x, type:x, connected }}
    if sub_feature == "intf":
        return format_intf(val_file, os_type, key, ntc_output)
    ### SWITCHPORT: {intf: {mode: access or trunk, vlan: x or [x,y]}}
    elif sub_feature == "switchport":
        return format_swport(val_file, ntc_output)
    ### IP_BRIEF: {intf: {ip:x, status: x}}}
    elif sub_feature == "ip_brief":
        return format_ipbrief(val_file, os_type, key, ntc_output)
    ### CatchAll
    else:
        msg = f"Unsupported sub_feature: {sub_feature}"
        raise ValueError(msg)



import ipaddress
import re
from collections import defaultdict
from typing import Any, NamedTuple


# ----------------------------------------------------------------------------
# KEY: Set dictionary keys on a per-os_type basis
# ----------------------------------------------------------------------------
class OsKeys(NamedTuple):
    count_match: str
    count_iter: int
    route_mask: str
    route_nhip: str
    route_nhif: str
    route_type: str


def _set_keys(os_type: str) -> OsKeys:
    """Based on the OS type set the set the dictionary keys when gleaning data from NTC data structures.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
    Returns:
        OsKeys: Dictionary Keys for the specific OS type to retrieve the output data
    """
    if "ios" in os_type:
        return OsKeys("IP", 2, "prefix_length", "nexthop_ip", "nexthop_if", "protocol")
    elif "nxos" in os_type:
        return OsKeys("IP", -1, "prefix_length", "nexthop_ip", "nexthop_if", "protocol")
    elif "asa" in os_type:
        return OsKeys("IP", 2, "netmask", "nexthopip", "nexthopif", "protocol")
    elif "panos" in os_type:
        return OsKeys("IPv6", 3, "prefix_length", "nexthop_ip", "nexthop_if", "flags")
    # Fallback if nothing matched
    msg = f"Error, '_set_keys' has no match for OS type: '{os_type}'"
    raise NotImplementedError(msg)


# ----------------------------------------------------------------------------
# OUTPUT: Creates str or ntc output dictionaries based on device command output
# ----------------------------------------------------------------------------
def _format_output(
    os_type: str, sub_feature: str, output: list[str | dict[str, str]]
) -> tuple[list[str], list[dict[str, str]]]:
    """Screen scraping return different data structures, they need defining to make function typing easier.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str from raw) command output from the device
    Raises:
        ValueError: Errors if the new output variable doesn't match the input as not meant to be changing it, just defining it for MYPY
    Returns:
        tuple[list[str], list[dict[str, str]]]: Returns either RAW list[str] or NTC list[dict], the other non-matched one will be empty
    """
    raw_output = [o for o in output if isinstance(o, str)]
    ntc_output = [o for o in output if isinstance(o, dict)]
    if (
        len(raw_output) != 0
        and output != raw_output
        or len(ntc_output) != 0
        and output != ntc_output
    ):
        msg = f"{os_type} {sub_feature} _format_output is malformed"
        raise ValueError(msg)
    return raw_output, ntc_output


# ----------------------------------------------------------------------------
# DEF: Mini-functions used by the main function
# ----------------------------------------------------------------------------
def _make_int(input_data: str) -> int | str:
    """Takes a string and returns an integer if it can, otherwise it returns the original string.

    Args:
        input_data (str): The data to be converted to an integer
    Returns:
        int | str: The input_data as a integer if possible, if not as the original string
    """
    try:
        return int(input_data)
    except ValueError:
        return input_data


def format_vrf(output: list[dict[str, Any]]) -> dict[str | int, Any]:
    """Format VRF output into the data structure.

    Args:
        output (list[dict[str, str]]): The command output from the device in ntc data structure
    Returns:
        dict[str, Any]: {vrf: [intfx, intfy]}
    """
    result: dict[str | int, list[str]] = {}
    for each_vrf in output:
        vrf_name = _make_int(each_vrf["name"])
        if result.get(vrf_name) is None:
            try:
                result[vrf_name] = each_vrf["interfaces"]
            except KeyError:
                result[vrf_name] = [each_vrf["interface"]]
        else:
            result[vrf_name].append(each_vrf["interface"])
    return dict(result)


def format_rte_count(key: OsKeys, os_type: str, output: list[str]) -> dict[str, Any]:
    """Format route count output into the data structure.

    Args:
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        os_type (str): The different Nornir platforms which are OS type of the device
        output (list[dict[str, str]]): The command output from the device in raw data structure
    Returns:
        dict[str, Any]: {global_subnets: x, xx_subnets: x}
    """
    result: dict[str, int | str] = {}
    for idx, each_item in enumerate(output):
        if key.count_match in each_item:
            # Needed for Panos as no VRFs
            if bool(re.search("panos", os_type)):
                vrf = "global"
            else:
                tmp = each_item.replace("maximum-paths is", "name is default")
                vrf = tmp.split()[5].replace('"', "").replace("default", "global")
            result[vrf] = _make_int(output[idx + 1].split()[key.count_iter])
    return dict(result)


def format_route(key: OsKeys, output: list[dict[str, Any]]) -> dict[str, Any]:
    """Format table output into a structured route dictionary.

    Args:
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        output (list[dict[str, Any]]): The command output from the device in ntc data structure

    Returns:
        dict[str, Any]: {vrf: {route/prefix: type: x, nh: y}})
    """
    result: dict[str, dict[str, dict[str, str | list[str]]]] = defaultdict(dict)

    # Helper functions used only by this function
    def _get_pfxlen(network: str, mask: str) -> str:
        """Takes a network and a mask and returns the network/prefix length."""
        ip_mask = network + "/" + mask
        ip_pfxlen = ipaddress.IPv4Interface(ip_mask).with_prefixlen
        return ip_pfxlen

    def _select_next_hop(each_rte: dict[str, str]) -> str:
        """Determine next-hop IP or interface."""
        proto_regex = re.compile(r"^L$|^local$|^C$|^direct$|A C")
        if proto_regex.search(each_rte[key.route_type]) or not each_rte[key.route_nhip]:
            return each_rte[key.route_nhif]
        return each_rte[key.route_nhip]

    def _format_route_type(each_rte: dict[str, str]) -> str:
        """Format route type string."""
        return (
            each_rte[key.route_type].replace("A ", "").replace("A?", "")
            if not each_rte.get("type", "")
            else f"{each_rte[key.route_type]} {each_rte['type']}"
        )

    for each_rte in output:
        if not isinstance(each_rte, dict):
            continue
        # VRF Handling
        vrf = each_rte.get("vrf", "global").replace("default", "global") or "global"
        # Route + Next-Hop
        rte = _get_pfxlen(each_rte["network"], each_rte[key.route_mask])
        nh = _select_next_hop(each_rte)
        rte_type = _format_route_type(each_rte)

        # Insert the above into result
        vrf_routes = result.setdefault(vrf, {})
        if rte not in vrf_routes:
            # First time seeing this route
            vrf_routes[rte] = {"nh": nh, "rtype": rte_type}
        else:
            existing_nh = vrf_routes[rte]["nh"]
            if isinstance(existing_nh, list):
                existing_nh.append(nh)
            elif existing_nh != nh:
                vrf_routes[rte]["nh"] = [existing_nh, nh]

    return dict(result)


# ----------------------------------------------------------------------------
# ACTUAL_STATE: Engine use to create sub-feature actual state or validation file
# ----------------------------------------------------------------------------
def format_actual_state(
    val_file: bool,  # noqa: ARG001
    os_type: str,
    sub_feature: str,
    output: list[str | dict[str, str]],
) -> dict[Any, Any]:
    """Engine to run all the actual state and validation file sub-feature formatting.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str/int from raw) command output from the device
    Returns:
        dict[str, Any]: Returns cmd output formatted into the data structure of actual state or validation file
    """
    key = _set_keys(os_type)
    raw_output, ntc_output = _format_output(os_type, sub_feature, output)

    ### VRF: {vrf: [intfx, intfy]}
    if sub_feature == "vrf":
        return format_vrf(ntc_output)

    ### RTE_COUNT: {global_subnets: x, xx_subnets: x}
    elif sub_feature == "route_count":
        return format_rte_count(key, os_type, raw_output)

    ### ROUTE: {vrf: {route/prefix: type: x, nh: y}})
    elif sub_feature == "route":
        return format_route(key, ntc_output)

    ### CatchAll
    else:
        msg = f"Unsupported sub_feature: {sub_feature}"
        raise ValueError(msg)



from importlib.metadata import PackageNotFoundError, version

from nornir_validate.core import (
    print_build_result,
    print_val_result,
    val_file_builder,
    validate,
)

try:
    __version__ = version("nornir_validate")
except PackageNotFoundError:
    # Package isn't installed yet (dev mode)
    __version__ = "0.0.0"

__all__ = ["validate", "print_val_result", "print_val_result", "print_build_result"]



import re
from collections import defaultdict
from typing import Any, NamedTuple


# ----------------------------------------------------------------------------
# KEY: Set dictionary keys on a per-os_type basis
# ----------------------------------------------------------------------------
class OsKeys(NamedTuple):
    ha_state_local: str
    ha_state_peer: str


def _set_keys(os_type: str) -> OsKeys:
    """Based on the OS type set the set the dictionary keys when gleaning data from NTC data structures.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
    Returns:
        OsKeys: Dictionary Keys for the specific OS type to retrieve the output data
    """
    if "ios" in os_type:
        return OsKeys("active_software_state", "standby_software_state")
    elif "asa" in os_type:
        return OsKeys("service_state", "service_state_mate")
    elif "wlc" in os_type:
        return OsKeys("local_state", "peer_state")
    elif "panos" in os_type:
        return OsKeys("state", "")
    # Fallback if nothing matched
    msg = f"Error, '_set_keys' has no match for OS type: '{os_type}'"
    raise NotImplementedError(msg)


# ----------------------------------------------------------------------------
# OUTPUT: Creates str or ntc output dictionaries based on device command output
# ----------------------------------------------------------------------------
def _format_output(
    os_type: str, sub_feature: str, output: list[str | dict[str, str]]
) -> tuple[list[str], list[dict[str, str]]]:
    """Screen scraping return different data structures, they need defining to make function typing easier.

    Args:
        os_type (str): A list of strings that are the OS types of the devices in the inventory
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str from raw) command output from the device
    Raises:
        ValueError: Errors if the new output variable doesn't match the input as not meant to be changing it, just defining it for MYPY
    Returns:
        tuple[list[str], list[dict[str, str]]]: Returns either RAW list[str] or NTC list[dict], the other non-matched one will be empty
    """
    raw_output = [o for o in output if isinstance(o, str)]
    ntc_output = [o for o in output if isinstance(o, dict)]
    if (
        len(raw_output) != 0
        and output != raw_output
        or len(ntc_output) != 0
        and output != ntc_output
    ):
        msg = f"{os_type} {sub_feature} _format_output is malformed"
        raise ValueError(msg)
    return raw_output, ntc_output


# ----------------------------------------------------------------------------
# DEF: Mini-functions used by the main function
# ----------------------------------------------------------------------------
def _make_int(input_data: str) -> int | str:
    """Takes a string and returns an integer if it can, otherwise it returns the original string.

    Args:
        input_data (str): The data to be converted to an integer
    Returns:
        int | str: The input_data as a integer if possible, if not as the original string
    """
    try:
        return int(input_data)
    except ValueError:
        return input_data


def format_ha_state(
    key: OsKeys, os_type: str, output: list[dict[str, Any]]
) -> dict[str, str]:
    """Format HA state into the data structure.

    Args:
        key (OsKeys): Keys for the specific OS type to retrieve the output data
        os_type (str): The different Nornir platforms which are OS type of the device
        output (list[dict[str, Any]]): The command output from the device in ntc data structure
    Returns:
        dict[str, str]: {local_state: x, peer_state: x}
    """
    result = {}
    if bool(re.search("asa", os_type)):
        result["local"] = output[0][key.ha_state_local][0]
        result["peer"] = output[0][key.ha_state_peer][0]
    elif bool(re.search("panos", os_type)):
        state = output[0][key.ha_state_local].split("-")[0]
        result["local"] = state
        result["peer"] = "standby" if state == "active" else "active"
    else:
        result["local"] = output[0][key.ha_state_local]
        result["peer"] = output[0][key.ha_state_peer]
    return dict(result)


def format_sw_stack(val_file: bool, output: list[dict[str, Any]]) -> dict[str, Any]:
    """Format switch stack into the data structure.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        output (list[dict[str, Any]]): The command output from the device in ntc data structure
    Returns:
        dict[str, Any]: {switchx: {priority: x, role: x, state: x}}, val)file doesn't have 'state: x'
    """
    result: dict[str, dict[str, str | int]] = defaultdict(dict)
    for each_swi in output:
        swi = f"switch{_make_int(each_swi['switch'])}"
        result[swi]["role"] = each_swi["role"]
        result[swi]["priority"] = _make_int(each_swi["priority"])
        result[swi]["role"] = each_swi["role"]
        # If creating actual state also needs state (is implicit in val file)
        if not val_file:
            result[swi]["state"] = each_swi["state"]
    return dict(result)


# ----------------------------------------------------------------------------
# ACTUAL_STATE: Engine use to create sub-feature actual state or validation file
# ----------------------------------------------------------------------------
def format_actual_state(
    val_file: bool,  # noqa: ARG001
    os_type: str,
    sub_feature: str,
    output: list[str | dict[str, str]],
) -> dict[str, Any]:
    """Engine to run all the actual state and validation file sub-feature formatting.

    Args:
        val_file (bool): Used to identify if creating validation file as sometimes need implicit values
        os_type (str): The different Nornir platforms which are OS type of the device
        sub_feature (str): The name of the sub-feature that is being validated
        output (list[str | dict[str, str]]): The structured (dict from NTC template) or unstructured (str/int from raw) command output from the device
    Returns:
        dict[str, Any]: Returns cmd output formatted into the data structure of actual state or validation file
    """
    key = _set_keys(os_type)
    raw_output, ntc_output = _format_output(os_type, sub_feature, output)

    ### HA_STATE: {local_state: x, peer_state: x}
    if sub_feature == "ha_state":
        return format_ha_state(key, os_type, ntc_output)
    ### SWI_STACK: show switch - {switchx: {priority: x, role: x, state: x}}, val_file doesn't have 'state: x'.
    elif sub_feature == "sw_stack":
        return format_sw_stack(val_file, ntc_output)
    ### CatchAll
    else:
        msg = f"Unsupported sub_feature: {sub_feature}"
        raise ValueError(msg)



import ast
import contextlib
import importlib
import json
import logging
import os
import re
from collections import defaultdict
from collections.abc import Callable
from importlib.resources import files
from types import ModuleType
from typing import Any

import yaml
from nornir.core import Nornir
from nornir.core.exceptions import NornirSubTaskError
from nornir.core.inventory import Host
from nornir.core.task import AggregatedResult, Result, Task
from nornir_jinja2.plugins.tasks import template_file  # type: ignore
from nornir_netmiko.tasks import netmiko_send_command  # type: ignore
from nornir_rich.functions import print_result  # type: ignore
from nornir_utils.plugins.tasks.files import write_file  # type: ignore

from .compliance_report import generate_validate_report

# Module-level cache
_loaded_modules: dict[str, Any] = {}


# ----------------------------------------------------------------------------
# IMPORT: Import actual_state modules required based on validations in input data
# ----------------------------------------------------------------------------
def import_actual_state_modules(feature: str) -> dict[str, ModuleType]:
    """Imports the actual_state.py modules for the feature used in the validation input file, caches results to avoid re-importing.

    Args:
        feature (str): Feature name used in file path and xx__actual_state.py

    Returns:
        dict[str, ModuleType]: Dictionary of {feature_name: imported_module}
    """
    # Only import if not already cached
    if feature not in _loaded_modules:
        module_path = (
            f"nornir_validate.feature_templates.{feature}.{feature}_actual_state"
        )
        try:
            _loaded_modules[feature] = importlib.import_module(module_path)
        except ImportError as e:
            print(f"❌ Could not import {module_path}: {e}")
        except Exception as e:
            print(f"❌ Error loading {module_path}: {e}")
    return _loaded_modules


# ----------------------------------------------------------------------------
#  OS_TYPE: Gets os_type from host_var OS types (platform) and then removes duplicates and None
# ----------------------------------------------------------------------------
def merge_os_types(host: Host) -> list[str]:
    """For a Nornir host object gathers the connection handlers name for the different connection plugins.

    Args:
        host (Host): Nornir inventory host object, holds the hosts attributes
    Returns:
        list[str]: List of the connection handlers for the different connection plugins
    """
    tmp_os_type: list[str | None] = []
    tmp_os_type.append(host.platform)
    tmp_os_type.append(host.get_connection_parameters("scrapli").platform)
    tmp_os_type.append(host.get_connection_parameters("netmiko").platform)
    tmp_os_type.append(host.get_connection_parameters("napalm").platform)
    set_os_type = set(tmp_os_type)
    os_type = [x for x in set_os_type if x is not None]
    os_type.sort()
    return os_type


# ----------------------------------------------------------------------------
# CRUNCH: Combines features, sub-features and template_path into a structured dictionary
# ----------------------------------------------------------------------------
def return_feature_desired_data(validations: dict[str, Any]) -> dict[str, Any]:
    """Reformats validation structure adding sub_features key and name of the jinja template.

    Args:
        validations (dict[str, Any]): The user defined validations that were gathered from the input file
    Returns:
        dict[str, Any]: Dictionary of all the data required for the desired state to be rendered
    """
    feat_desired_data: dict[str, dict[str, Any]] = defaultdict(dict)
    for feature, sub_feature in validations.items():
        feat_desired_data[feature]["file"] = f"{feature}_desired_state.j2"
        if isinstance(sub_feature, list):
            feat_desired_data[feature]["sub_features"] = sub_feature
        else:
            feat_desired_data[feature]["sub_features"] = {}
            feat_desired_data[feature]["sub_features"].update(sub_feature)
    return feat_desired_data


# ----------------------------------------------------------------------------
#  CRUNCH: Converts the yaml formatted string into a dictionary
# ----------------------------------------------------------------------------
def return_yaml_desired_state(str_desired_state: str) -> dict[str, Any]:
    """Takes a string of YAML and returns a dictionary of the desired state.

    Args:
        str_desired_state (str): String representation of the desired state (including cmds)

    Returns:
        dict[str, Any]: A dictionary of the desired state in the format {feat: {subfeat: {cmds: expected_result}
    """
    # Conditional fix as yaml.Loader causes error for the pattern ">dd"
    if re.search(r" >\d+\n", str_desired_state):
        x = yaml.load(str_desired_state.replace(">", "->"), Loader=yaml.Loader)
        desired_state = ast.literal_eval(str(x[0]).replace("->", ">"))
    else:
        desired_state = yaml.load(str_desired_state, Loader=yaml.Loader)[0]
    return dict(desired_state)


# ----------------------------------------------------------------------------
#  CLEAN: Removes any features, sub-features or commands with a value of None
# ----------------------------------------------------------------------------
def strip_empty_feat(desired_state: dict[str, Any]) -> dict[str, Any]:
    """Removes any features, sub-features or commands with a value of None from desired state to safeguard against errors.

    Args:
        desired_state (dict[str, Any]): Desired state in format
    Returns:
        dict[str, Any]: Final desired state
    """
    # Remove any features that dont hold any sub features (not a dictionary)
    desired_state = {k: v for k, v in desired_state.items() if isinstance(v, dict)}
    # Remove any sub-features that dont hold any commands (not a dictionary)
    tmp_ds = {}
    for feat, sub_feat in desired_state.items():
        tmp_ds[feat] = {k: v for k, v in sub_feat.items() if isinstance(v, dict)}
    # Delete any commands that have a desired state of None
    clean_desired_state: dict[str, dict[str, Any]] = defaultdict(dict)
    for feat, sub_feat in tmp_ds.items():
        for sub_feat_name, cmds in sub_feat.items():
            tmp_cmd = {}
            for each_cmd, cmd_ds in cmds.items():
                if cmd_ds is not None and cmd_ds != "None":
                    tmp_cmd[each_cmd] = cmd_ds
            if len(tmp_cmd) != 0:
                clean_desired_state[feat][sub_feat_name] = tmp_cmd
    return dict(clean_desired_state)


# ----------------------------------------------------------------------------
#  REMOVE: Removes commands from the desired state sub-features
# ----------------------------------------------------------------------------
def remove_cmds_desired_state(desired_state: dict[str, Any]) -> dict[str, Any]:
    """Strips the commands out of the desired state dictionaries (no longer needed as gathered actual state).

    Args:
        desired_state (dict[str, Any]): Desired state in format ({feat: {subfeat: {cmd: expected_result})

    Returns:
        dict[str, Any]: Final desired state in the format {feat: {subfeat: expected_result}}
    """
    clean_desired_state: dict[str, dict[str, Any]] = defaultdict(dict)
    for feature, sub_feature in desired_state.items():
        for sub_feat_name, sub_feat_cmds in sub_feature.items():
            clean_desired_state[feature][sub_feat_name] = {}
            for cmd_ds in sub_feat_cmds.values():
                if cmd_ds == "SUB_FEATURE_COMBINED_CMD":
                    pass
                elif isinstance(cmd_ds, dict):
                    clean_desired_state[feature][sub_feat_name].update(cmd_ds)
                else:
                    clean_desired_state[feature][sub_feat_name] = cmd_ds
    return dict(clean_desired_state)


# ----------------------------------------------------------------------------
#  VAL_DM: Creates validation data model for all feature/sub-feature validations
# ----------------------------------------------------------------------------
def create_val_dm() -> dict[str, dict[str, list[str]]]:
    """Creates validation data model from all_subfeat_index file.

    Returns:
        dict[str, Any]: Validation DM with any example data (dicts) such as route table VRF names removed
    """
    all_index_file = files("nornir_validate").joinpath("index_files", "all_index.yml")
    validations: dict[str, dict[str, list[str]]]
    with all_index_file.open("r") as tmp_data:
        validations = yaml.load(tmp_data, Loader=yaml.FullLoader)
        for feat in validations["all"]:
            for idx, sub_feat in enumerate(validations["all"][feat]):
                if isinstance(sub_feat, dict):
                    validations["all"][feat][idx] = list(sub_feat.keys())[0]
    return validations


# ----------------------------------------------------------------------------
# 1. DESIRED_STATE: Create a host_var of desired state by running (per-os-type) task_template
# ----------------------------------------------------------------------------
def task_desired_state(
    task: Task,
    validations: dict[str, Any],
    task_template: Callable[[Task, str, str, dict[str, Any]], str],
) -> Result | None:
    """Uses nornir-template (task_template method) to produce host_var of combined desired states (exits if nothing to be validated).

    Args:
        task (Task): The nornir tasks that implements (runs) this the nornir-jinja template plugin
        validations (dict[str, Any]): The validations to be run (user input validation data)
        task_template (Callable[[Task, str, str, dict[str, Any]], str]): Nornir-jinja plugin task to render the validations
    Returns:
        Optional[Result]: If no validation returns Nornir result stating so (if validations desired_sate added as hosts_vars to task host
    """
    desired_state: dict[str, Any] = {}
    # 1a. TMPL: Create the desired_state for each feature to be validated (double logic to stop error if top level dict not exist)
    if (
        validations.get("hosts") is not None
        and validations["hosts"].get(str(task.host)) is not None
    ):
        task.run(
            task=task_template,
            tmpl_path="feature_templates/",
            validations=validations["hosts"][str(task.host)],
            desired_state=desired_state,
        )
    if (
        validations.get("groups") is not None
        and validations["groups"].get(str(task.host.groups[0])) is not None
    ):
        task.run(
            task=task_template,
            tmpl_path="feature_templates/",
            validations=validations["groups"][str(task.host.groups[0])],
            desired_state=desired_state,
        )
    if validations.get("all") is not None:
        task.run(
            task=task_template,
            tmpl_path="feature_templates/",
            validations=validations["all"],
            desired_state=desired_state,
        )
    # 1b. VAR: Create host_var of combined desired states or exits if nothing to be validated

    if len(desired_state) == 0:
        result_text = "\u26a0\ufe0f  No validations were performed as no desired_state was generated, check input file and template"
        return Result(host=task.host, failed=True, result=result_text)
    else:
        task.host["desired_state"] = strip_empty_feat(desired_state)
        return None


# ----------------------------------------------------------------------------
# 2. TEMPLATE: Called by task_desired_state to create the desired state (using jinja)
# ----------------------------------------------------------------------------
def task_template(
    task: Task,
    tmpl_path: str,
    validations: dict[str, Any],
    desired_state: dict[str, Any],
) -> dict[str, Any]:
    """Nornir-jinja task to render the validations into the desired states (also includes cmds to get the actual-state).

    Args:
        task (Task): The nornir tasks that implements (runs) this the nornir-jinja template plugin
        tmpl_path (str): Path to the feature templates
        validations (dict[str, Any]): The validations to be run (user input validation data)
        desired_state (dict[str, Any]): The final desired state in format ({feat: {subfeat: {cmd: expected_result}).

    Returns:
        list[str]: _description_

    # :param task: The task object that is passed to the plugin
    # :type task: Task
    """
    # 2a. CRUNCH: Formulate data to be used in templates to create desired state
    os_type = merge_os_types(task.host)
    feat_desired_data = return_feature_desired_data(validations)
    # 2b. TMPL: Create the desired state (includes cmds to get actual state) from the jinja2 template
    for feature, values in feat_desired_data.items():
        str_desired_state: str = task.run(
            task=template_file,
            template=values["file"],
            path=files("nornir_validate").joinpath(tmpl_path, feature),
            os_type=os_type,
            feature=feature,
            sub_features=values["sub_features"],
        ).result
        # 2c. SERIALISE: Uses YAML to convert the string result into a dictionary ({feat: {subfeat: {cmd: expected_result})
        desired_state.update(return_yaml_desired_state(str_desired_state))

    return desired_state


# ----------------------------------------------------------------------------
# 3. ACTUAL_STATE: Formats cmd outputs to create the actual state
# ----------------------------------------------------------------------------
def actual_state_engine(
    val_file: bool, os_type: list[str], feat_actual_data: dict[str, dict[str, Any]]
) -> dict[str, dict[str, Any]]:
    """From the cmd output creates the actual state of features and sub-features with the output of the sub-features formatted.

    Args:
        val_file (bool): True if generate validation file called this function
        os_type (list[str]): Connection handler (plugin) used to format cmd data into actual state structure (same format as desired state)
        feat_actual_data (dict[str, dict[str, Any]]): The structured or non-structured data (cmd output) got from devices
    Returns:
        dict[str, dict[str, Any]]: Actual state formatted as ({feat: {subfeat: actual_result})
    """
    actual_state: dict[str, dict[str, Any]] = defaultdict(dict)
    for feature, sub_feat_dict in feat_actual_data.items():
        for sub_feature, output in sub_feat_dict.items():
            # EMPTY: If output is empty just adds an empty dictionary
            if output is None or len(output) == 0:
                result = {}
            else:
                # Gets per-sub-feature actual state structured data from imported feature_templates (python imports)
                modules = import_actual_state_modules(feature)
                result = modules[feature].format_actual_state(
                    val_file, str(os_type), sub_feature, output
                )
            actual_state[feature][sub_feature] = result
    return dict(actual_state)


# ----------------------------------------------------------------------------
# 4. ENGINE: Formats gathered output as actual state and runs compliance report - Only one that prints (logging debug)
# ----------------------------------------------------------------------------
def validate(
    task: Task, input_data: dict[str, Any], save_report: str | None = None
) -> Result:
    """The main engine that runs file formatting, nornir tasks and compliance report.

    Args:
        task (Task): The nornir tasks that implements (runs) this the nornir tasks
        input_data (str): The User defined input data from input file
        save_report (str | None): To optionally save compliance reports to the directory specified in this variable
    Returns:
        Result: The final result of nornir_validate (all tasks), a special Nornir Result object passed back to the main() method to be printed
    """
    # 4a. TMPL: Creates desired states using the jinja template by calling task_desired_state (1) which in term calls task_template (2)
    task.run(
        task=task_desired_state,
        validations=input_data,
        task_template=task_template,
        severity_level=logging.DEBUG,
    )
    # 4b. CMD: Using commands crunched from the desired output gathers per-feature/sub-feature actual config of the device
    feat_actual_data: dict[str, dict[str, Any]] = defaultdict(dict)
    for feature, sub_feature in task.host["desired_state"].items():
        for sub_feat_name, sub_feat_cmds in sub_feature.items():
            cmd_output = []
            for cmd in sub_feat_cmds.keys():  # noqa: SIM118
                tmp_cmd_output = task.run(
                    task=netmiko_send_command,
                    command_string=cmd,
                    use_textfsm=True,
                    severity_level=logging.DEBUG,
                ).result
                # Converts NXOS "| json" cmds from string to JSON (suppress exceptions in case feature not used)
                if "json" in cmd:
                    with contextlib.suppress(Exception):
                        tmp_cmd_output = [json.loads(tmp_cmd_output)]
                # Required for non-structured data(no NTC template)
                elif isinstance(tmp_cmd_output, str):
                    tmp_cmd_output = tmp_cmd_output.lstrip().rstrip().splitlines()
                cmd_output.extend(tmp_cmd_output)
            feat_actual_data[feature][sub_feat_name] = cmd_output

    # 4c. ACTUAL: Formats the returned data into dict of cmds {cmd: {seq: key:val}} same as desired_state
    os_type = merge_os_types(task.host)
    actual_state = actual_state_engine(False, os_type, feat_actual_data)
    # 4d. VAL: Uses Napalm_validate validate method to generate a compliance report
    desired_state = remove_cmds_desired_state(task.host["desired_state"])
    comp_result = generate_validate_report(
        desired_state, actual_state, str(task.host), save_report
    )
    # 4e. RSLT: Nornir returns compliance result or if fails the compliance report
    return Result(
        host=task.host,
        failed=comp_result["failed"],
        result=comp_result["result"],
        report=comp_result["report"],
        report_text=comp_result["report_text"],
    )


# ----------------------------------------------------------------------------
# 5. VAL_FILE_BUILDER: Builds validation files based on the actual state
# ----------------------------------------------------------------------------
def val_file_builder(
    task: Task, input_data: dict[str, Any] | str = "", directory: str = ""
) -> Result:
    """Generates a validation file based on what features are enabled on a device (gathered from actual state).

    Args:
        task (Task): The Nornir task that executes host actions and stores the results
        input_data (dict[str, Any] | str): Validations or if an empty string if dynamically creating a validation file
        directory (str): Working directory where the file will be saved
    Returns:
        Result: The nornir result from the execution of the task, so list of enabled and not enabled features as well as val file name
    """
    # ERROR: Error patterns returned by devices that would mean a feature is not configured
    error_patterns = [
        "ERROR: % Invalid input detected at '^' marker.",
        "% Invalid input detected at '^' marker.",
        "Invalid command at '^' marker.",
    ]
    # 5a. Use input input validation DM or if empty create one of all the possible validations
    validations = create_val_dm() if len(input_data) == 0 else input_data
    # 5b. TMPL: Creates desired states using the jinja template by calling task_desired_state (1) which in term calls task)template (2)
    task.run(
        task=task_desired_state,
        validations=validations,
        task_template=task_template,
        severity_level=logging.DEBUG,
    )
    # 5c. CMD: Using commands crunched from the desired output gathers pre-feature/sub-feature actual config of the device
    feat_actual_data: dict[str, dict[str, Any]] = defaultdict(dict)
    used_subfeat, not_used_subfeat = ([] for i in range(2))
    for feature, sub_feature in task.host["desired_state"].items():
        for sub_feat_name, sub_feat_cmds in sub_feature.items():
            cmd_output = []
            for cmd in sub_feat_cmds.keys():  # noqa: SIM118
                try:
                    tmp_cmd_output = task.run(
                        task=netmiko_send_command,
                        command_string=cmd,
                        use_textfsm=True,
                        severity_level=logging.DEBUG,
                    ).result
                except NornirSubTaskError:
                    tmp_cmd_output = []
                # Converts NXOS "| json" cmds from string to JSON (suppress exceptions in case feature not used)
                if "json" in cmd:
                    with contextlib.suppress(Exception):
                        tmp_cmd_output = [json.loads(tmp_cmd_output)]
                # Required for non-structured data(no NTC template)
                elif isinstance(tmp_cmd_output, str):
                    tmp_cmd_output = tmp_cmd_output.lstrip().rstrip().splitlines()
                cmd_output.extend(tmp_cmd_output)
            if len(tmp_cmd_output) != 0:
                # SKIP: Skips sub-feature validation if command returned an error
                if any(pattern in str(tmp_cmd_output) for pattern in error_patterns):
                    not_used_subfeat.append(sub_feat_name)
                else:
                    feat_actual_data[feature][sub_feat_name] = cmd_output
                    used_subfeat.append(sub_feat_name)
            else:
                not_used_subfeat.append(sub_feat_name)

    #  5d. FORMAT: Format the returned data into dict of cmds {cmd: {seq: key:val}} and save to file
    os_type = merge_os_types(task.host)
    actual_state = actual_state_engine(True, os_type, feat_actual_data)
    val_file = os.path.join(directory, f"{str(task.host)}_vals.yml")
    task.run(
        task=write_file,
        filename=val_file,
        content=yaml.dump({"hosts": {str(task.host): actual_state}}, sort_keys=False),
    )
    info = f"✅ Created the validation file '{val_file}'"
    return Result(
        host=task.host,
        result="",
        used_subfeat=used_subfeat,
        not_used_subfeat=not_used_subfeat,
        file_info=info,
    )


# ----------------------------------------------------------------------------
# 6. PRINT: Functions for printing results
# ----------------------------------------------------------------------------
def print_val_result(result: AggregatedResult) -> None:
    """Prints the result of the validation in a user friendly way using a customization of nornir_rich print (better handling of dicts).

    Args:
        result (AggregatedResult): The nornir result from the execution of the task, so the compliance report in Nornir Result format
    """
    print_result(result)


def print_build_result(result: AggregatedResult, nr: Nornir) -> None:
    """Prints the result of the validation file builder, required in this manner so that failed tasks (features not enabled) dont show as a failed result.

    Args:
        result (AggregatedResult): The nornir result from the execution of the task, am only printing results for specified vars
        nr (Nornir,): The Nornir inventory from which get (loop through) hostnames
    """
    for each_host in nr.inventory.hosts.keys():  # noqa: SIM118
        print_result(
            result[each_host][0],
            vars=[
                "host",
                "result",
                "used_subfeat",
                "not_used_subfeat",
                "file_info",
            ],
        )
